"""Commands to check for workflow and task status (from CLI)."""

import getpass
import logging
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from jobmon.client.logging import JobmonLoggerConfig
from jobmon.client.swarm.workflow_run import WorkflowRun as SwarmWorkflowRun
from jobmon.client.workflow import DistributorContext
from jobmon.client.workflow_run import WorkflowRunFactory
from jobmon.core.constants import (
    ExecludeTTVs,
    TaskStatus,
    WorkflowRunStatus,
    WorkflowStatus,
)
from jobmon.core.exceptions import WorkflowRunStateError
from jobmon.core.requester import Requester
from jobmon.core.serializers import SerializeTaskTemplateResourceUsage


logger = logging.getLogger(__name__)


def workflow_status(
    workflow_id: Optional[List[int]] = None,
    user: Optional[List[str]] = None,
    json: bool = False,
    limit: Optional[int] = 5,
    requester: Optional[Requester] = None,
) -> pd.DataFrame:
    """Get metadata about workflow progress.

    Args:
        workflow_id: workflow_id/s to retrieve info for. If not specified will pull all
            workflows by user
        user: user/s to retrieve info for. If not specified will return for current user.
        limit: return # of records order by wf id desc. Return 5 if not provided;
            return all if [], [<0].
        json: Flag to return data as JSON
        requester: object to communicate with the flask services

    Returns:
        dataframe of all workflows and their status
    """
    if workflow_id is None:
        workflow_id = []
    if user is None:
        user = []
    logger.debug("workflow_status workflow_id:{}".format(str(workflow_id)))
    msg: dict = {}
    if workflow_id:
        msg["workflow_id"] = workflow_id
    if user:
        msg["user"] = user
    else:
        msg["user"] = getpass.getuser()
    msg["limit"] = limit

    if requester is None:
        requester = Requester.from_defaults()

    rc, res = requester.send_request(
        app_route="/workflow_status", message=msg, request_type="get"
    )
    if json:
        return res["workflows"]
    else:
        df = pd.read_json(res["workflows"])
        # Cast CREATED_DATE back to a date-like object, serialized as an int
        df["CREATED_DATE"] = pd.to_datetime(df["CREATED_DATE"], unit="ms")
        return df


def workflow_tasks(
    workflow_id: int,
    status: Optional[List[str]] = None,
    json: bool = False,
    limit: int = 5,
    requester: Optional[Requester] = None,
) -> pd.DataFrame:
    """Get metadata about task state for a given workflow.

    Args:
        workflow_id: workflow_id/s to retrieve info for
        status: limit task state to one of [PENDING, RUNNING, DONE, FATAL] tasks
        json: Flag to return data as JSON
        limit: return # of records order by wf id desc. Return 5 if not provided
        requester: object to communicate with the flask services

    Returns:
        Dataframe of tasks for a given workflow
    """
    logger.debug("workflow id: {}".format(workflow_id))
    msg: Dict[str, Union[List[str], int]] = {}
    if status:
        msg["status"] = [i.upper() for i in status]
    msg["limit"] = limit

    if requester is None:
        requester = Requester.from_defaults()

    rc, res = requester.send_request(
        app_route=f"/workflow/{workflow_id}/workflow_tasks",
        message=msg,
        request_type="get",
    )
    if json:
        return res["workflow_tasks"]
    else:
        return pd.read_json(res["workflow_tasks"])


def task_template_resources(
    task_template_version: int,
    workflows: Optional[list] = None,
    node_args: Optional[Dict] = None,
    ci: Optional[float] = None,
    requester: Optional[Requester] = None,
) -> Optional[Dict]:
    """Get aggregate resource usage data for a given TaskTemplateVersion.

    Args:
        task_template_version: The task template version ID the user wants to find the
            resource usage of.
        workflows: list of workflows a user wants query by.
        node_args: dictionary of node arguments a user wants to query by.
        ci: confidence interval. Not calculate if None.
        requester: object to communicate with the flask services

    Returns:
        Dataframe of TaskTemplate resource usage
    """
    execlue_list = ExecludeTTVs.EXECLUDE_TTVS
    if task_template_version in execlue_list:
        msg = (
            f"Resource usage query for task_template_version {task_template_version}"
            f"  is restricted due to excessive size."
        )
        logger.warning(msg)
        print(msg)
        return None
    message: Dict[Any, Any] = dict()
    message["task_template_version_id"] = task_template_version
    if workflows:
        message["workflows"] = workflows
    if node_args:
        message["node_args"] = node_args
    if ci:
        message["ci"] = ci

    if requester is None:
        requester = Requester.from_defaults()

    app_route = "/task_template_resource_usage"
    return_code, response = requester.send_request(
        app_route=app_route, message=message, request_type="post"
    )

    def format_bytes(value: Any) -> Optional[str]:
        if value is not None:
            return str(value) + "B"
        else:
            return value

    kwargs = SerializeTaskTemplateResourceUsage.kwargs_from_wire(response)
    resources = {
        "num_tasks": kwargs["num_tasks"],
        "min_mem": format_bytes(kwargs["min_mem"]),
        "max_mem": format_bytes(kwargs["max_mem"]),
        "mean_mem": format_bytes(kwargs["mean_mem"]),
        "min_runtime": kwargs["min_runtime"],
        "max_runtime": kwargs["max_runtime"],
        "mean_runtime": kwargs["mean_runtime"],
        "median_mem": format_bytes(kwargs["median_mem"]),
        "median_runtime": kwargs["median_runtime"],
        "ci_mem": kwargs["ci_mem"],
        "ci_runtime": kwargs["ci_runtime"],
    }

    return resources


def task_status(
    task_ids: List[int],
    status: Optional[List[str]] = None,
    json: bool = False,
    requester: Optional[Requester] = None,
) -> Union[dict, pd.DataFrame]:
    """Get metadata about a task and its task instances.

    Args:
        task_ids: a list of task_ids to retrieve task_instance metadata for.
        status: a list of statuses to check for.
        json: Flag to return data as JSON.
        requester: object to communicate with the flask services

    Returns:
        Task status and task_instance metadata
    """
    logger.debug("task_status task_ids:{}".format(str(task_ids)))
    msg: Dict[str, Union[List[str], List[int]]] = {"task_ids": task_ids}
    if status:
        msg["status"] = [i.upper() for i in status]

    if requester is None:
        requester = Requester.from_defaults()

    rc, res = requester.send_request(
        app_route="/task_status", message=msg, request_type="get"
    )
    if json:
        return res["task_instance_status"]
    else:
        return pd.read_json(res["task_instance_status"], dtype=False)


def concurrency_limit(
    workflow_id: int,
    max_tasks: int,
    requester: Optional[Requester] = None,
) -> str:
    """Update a workflow's max_concurrently_running field in the database.

    Used to dynamically adjust the allowed number of jobs concurrently running.

    Args:
        workflow_id (int): ID of the running workflow whose max_running value needs to be reset
        max_tasks (int) : new allowed value of parallel tasks
        requester: object to communicate with the flask services

    Returns: string displaying success or failure of the update.
    """
    msg = {"max_tasks": max_tasks}

    if requester is None:
        requester = Requester.from_defaults()

    _, resp = requester.send_request(
        app_route=f"/workflow/{workflow_id}/update_max_concurrently_running",
        message=msg,
        request_type="put",
    )

    return resp["message"]


def _chunk_ids(ids: List[int], chunk_size: int = 100) -> List[List[int]]:
    """Chunk the ids into a list of 100 ids list.

    Args:
        ids: list of ids
        chunk_size: the size of each chunk; default to 100

    Returns: a list of list

    """
    return_list = []
    return_list.append(ids[0 : min(chunk_size, len(ids))])
    i = 1
    while i * chunk_size < len(ids):
        return_list.append(ids[i * chunk_size : min((i + 1) * chunk_size, len(ids))])
        i += 1
    return return_list


def update_task_status(
    task_ids: List[int],
    workflow_id: int,
    new_status: str,
    force: bool = False,
    recursive: bool = False,
    requester: Optional[Requester] = None,
) -> Any:
    """Set the specified task IDs to the new status, pending validation.

    Args:
        task_ids: List of task IDs to reset in the database
        workflow_id: The workflow to which each task belongs. Users can only self-service
            1 workflow at a time for the moment.
        new_status: the status to set tasks to
        force: if true, allow all source statuses and all workflow statuses.
        recursive: if true and force, apply recursive update_status downstream
            or upstream depending on new_status
            (upstream if new_status == 'D'; downstream if new_status == 'G').
        requester: object to communicate with the flask services
    """
    if requester is None:
        requester = Requester.from_defaults()

    # Validate the username is appropriate
    user = getpass.getuser()

    validate_username(workflow_id, user, requester)
    workflow_status = validate_workflow(task_ids, requester)

    # Validate the allowed statuses. For now, only "D" and "G" allowed.
    allowed_statuses = [TaskStatus.REGISTERING, TaskStatus.DONE]
    assert (
        new_status in allowed_statuses
    ), f"Only {allowed_statuses} allowed to be set via CLI"

    # Conditional logic: If the new status is "D", only need to set task to "D"
    # Else: All downstreams must also be set to "G", and task instances set to "K"
    if force and recursive:
        rc, res = requester.send_request(
            app_route="/tasks_recursive/" + ("up" if new_status == "D" else "down"),
            message={"task_ids": task_ids},
            request_type="put",
        )
        if rc != 200:
            raise AssertionError(f"Server return HTTP error code: {rc}")
        task_ids = res["task_ids"]
    else:
        if new_status == TaskStatus.REGISTERING:
            subdag_tasks = get_sub_task_tree(task_ids, requester=requester).keys()
            task_ids = task_ids + [*subdag_tasks]

    # We want to prevent excessive requests, with a hard-limit of 10,000 set up
    # to avoid churning on the server.
    if len(task_ids) > 10_000:
        raise AssertionError(
            f"There are too many tasks ({len(task_ids)}) requested "
            f"for the update. Request denied."
        )

    _, resp = requester.send_request(
        app_route="/task/update_statuses",
        message={
            "task_ids": task_ids,
            "new_status": new_status,
            "workflow_status": workflow_status,
            "workflow_id": workflow_id,
        },
        request_type="put",
    )

    return resp


def validate_username(workflow_id: int, username: str, requester: Requester) -> None:
    """Validate that the user is approved to make these changes."""
    rc, res = requester.send_request(
        app_route=f"/workflow/{workflow_id}/validate_username/{username}",
        message={},
        request_type="get",
    )
    if not res["validation"]:
        raise AssertionError(f"User {username} is not allowed to reset this workflow.")
    return


def validate_workflow(task_ids: List[int], requester: Requester) -> WorkflowStatus:
    """Validate workflow.

    The task_ids provided belong to the expected workflow,
    and the workflow status is in expected status unless we want to force
    it through.
    """
    rc, res = requester.send_request(
        app_route="/workflow_validation",
        message={"task_ids": task_ids},
        request_type="post",
    )

    if not bool(res["validation"]):
        raise AssertionError(
            "The workflow status of the given task ids are out of "
            "scope of the following required statuses "
            "(FAILED, DONE, ABORTED, HALTED) or multiple workflow statuses "
            "were found."
        )
    return res["workflow_status"]


def get_sub_task_tree(
    task_ids: list,
    task_status: Optional[list] = None,
    requester: Optional[Requester] = None,
) -> dict:
    """Get the sub_tree from tasks to ensure that they end up in the right states."""
    # This is to make the test case happy. Otherwise, requester should not be None.
    if requester is None:
        requester = Requester.from_defaults()

    # Valid input
    rc, res = requester.send_request(
        app_route="/task/subdag",
        message={"task_ids": task_ids, "task_status": task_status},
        request_type="post",
    )
    if rc != 200:
        raise AssertionError(f"Server return HTTP error code: {rc}")
    task_tree_dict = res["sub_task"]
    return task_tree_dict


def get_task_dependencies(task_id: int, requester: Optional[Requester] = None) -> dict:
    """Get the upstream and down stream of a task."""
    # This is to make the test case happy. Otherwise, requester should not be None.
    if requester is None:
        requester = Requester.from_defaults()
    # Valid input
    rc, res = requester.send_request(
        app_route=f"/task_dependencies/{task_id}", message={}, request_type="get"
    )
    if rc != 200:
        if rc == 404:
            raise AssertionError(
                f"Server return HTTP error code: {rc}. "
                f"The jobmon server version may not support this command."
            )
        else:
            raise AssertionError(f"Server return HTTP error code: {rc}")
    return res


def workflow_reset(
    workflow_id: int, partial_reset: bool = False, requester: Optional[Requester] = None
) -> str:
    """Workflow reset.

    Return:
        A string to indicate the workflow_reset result.

    Args:
        workflow_id: the workflow id to be reset.
        partial_reset: if False, tasks in D state will be reset as well
        requester: http server interface.
    """
    if requester is None:
        requester = Requester.from_defaults()

    username = getpass.getuser()

    rc, res = requester.send_request(
        app_route=f"/workflow/{workflow_id}/validate_for_workflow_reset/{username}",
        message={},
        request_type="get",
    )
    if res["workflow_run_id"]:
        # Terminate a workflow's active workflowruns. Should go to a terminated state
        rc, _ = requester.send_request(
            app_route=f"/workflow/{workflow_id}/reset",
            message={"partial_reset": partial_reset},
            request_type="put",
        )
        wr_return = f"Workflow {workflow_id} has been reset."
    else:
        wr_return = (
            f"User {username} is not the latest initiator of "
            f"workflow {workflow_id} that has resulted in error('E' status). "
            f"The workflow {workflow_id} has not been reset."
        )

    return wr_return


def _get_yaml_data(
    wfid: int, tid: int, v_mem: str, v_core: str, v_runtime: str, requester: Requester
) -> Dict:
    # make it a method for easy mock
    tt_exclude_list = ExecludeTTVs.EXECLUDE_TTVS

    key_map_m = {"avg": "mean_mem", "min": "min_mem", "max": "max_mem"}
    key_map_r = {"avg": "mean_runtime", "min": "min_runtime", "max": "max_runtime"}

    # Get task template version ids
    rc, res = requester.send_request(
        app_route="/get_task_template_version",
        message={"task_id": tid} if wfid is None else {"workflow_id": wfid},
        request_type="get",
    )
    if rc != 200:
        raise AssertionError(
            f"Server returns HTTP error code: {rc} " f"for get_task_template_version."
        )

    # data structure: {ttv_id: [name, core, mem, runtime, queue]}
    ttvis_dic = dict()

    # create a holder for tts that are in the exclude list
    tt_in_exclude = []
    for t in res["task_template_version_ids"]:
        if t["id"] in tt_exclude_list:
            tt_in_exclude.append(t)
            msg = f"Task template {t} contains too many tasks and will be excluded."
            logger.warning(msg)
            print(msg)
        else:
            ttvis_dic[t["id"]] = [t["name"]]

    for t in tt_in_exclude:
        # fill in template using default values
        print(f"Fill template for t using {t}")
        ttvis_dic[t["id"]] = [t["name"], 1, 1, 3600, "all.q"]

    # set of ttvs not in exclude list
    actual_ttvs = set(ttvis_dic.keys()) - set([i["id"] for i in tt_in_exclude])

    if len(actual_ttvs) == 0:
        return ttvis_dic

    # get core
    ttvis = str([i for i in actual_ttvs]).replace("[", "(").replace("]", ")")
    rc, res = requester.send_request(
        app_route="/get_requested_cores",
        message={"task_template_version_ids": f"{ttvis}"},
        request_type="get",
    )
    if rc != 200:
        raise AssertionError(
            f"Server returns HTTP error code: {rc} " f"for /get_requested_cores."
        )
    core_info = res["core_info"]
    for record in core_info:
        ttvis_dic[int(record["id"])].append(record[v_core])

    # Get actually mem and runtime for each ttvi
    for ttv in actual_ttvs:
        rc, res = requester.send_request(
            app_route="/task_template_resource_usage",
            message={"task_template_version_id": ttv},
            request_type="post",
        )
        if rc != 200:
            raise AssertionError(
                f"Server returns HTTP error code: {rc} "
                f"for /task_template_resource_usage."
            )
        usage = SerializeTaskTemplateResourceUsage.kwargs_from_wire(res)
        ttvis_dic[int(ttv)].append(int(usage[key_map_m[v_mem]]))
        ttvis_dic[int(ttv)].append(int(usage[key_map_r[v_runtime]]))

    # get queue
    rc, res = requester.send_request(
        app_route="/get_most_popular_queue",
        message={"task_template_version_ids": f"{ttvis}"},
        request_type="get",
    )
    if rc != 200:
        raise AssertionError(
            f"Server returns HTTP error code: {rc} " f"for /get_most_popular_queue."
        )
    for record in res["queue_info"]:
        ttvis_dic[int(record["id"])].append(record["queue"])
    return ttvis_dic


def _create_yaml(data: Optional[Dict] = None, clusters: Optional[List] = None) -> str:
    yaml = "task_template_resources:\n"
    if clusters is None:
        clusters = []
    if data is None or len(clusters) == 0:
        return yaml
    for ttv in data.keys():
        yaml += f"  {data[ttv][0]}:\n"  # name
        for cluster in clusters:
            yaml += f"    {cluster}:\n"  # cluster
            yaml += f"      cores: {data[ttv][1]}\n"  # core
            yaml += f'      memory: "{data[ttv][2]}B"\n'  # mem
            yaml += f"      runtime: {data[ttv][3]}\n"  # runtime
            yaml += f'      queue: "{data[ttv][4]}"\n'  # queue
    return yaml


def create_resource_yaml(
    wfid: int,
    tid: int,
    v_mem: str,
    v_core: str,
    v_runtime: str,
    clusters: List,
    requester: Optional[Requester] = None,
) -> str:
    """The method to create resource yaml."""
    if requester is None:
        requester = Requester.from_defaults()

    ttvis_dic = _get_yaml_data(wfid, tid, v_mem, v_core, v_runtime, requester)
    yaml = _create_yaml(ttvis_dic, clusters)
    return yaml


def get_filepaths(
    workflow_id: int,
    array_name: str = "",
    job_name: str = "",
    limit: int = 5,
    requester: Optional[Requester] = None,
) -> dict:
    """Allows users to get the stdout/stderr paths of their tasks."""
    if requester is None:
        requester = Requester.from_defaults()

    app_route = f"/array/{workflow_id}/get_array_tasks"
    _, resp = requester.send_request(
        app_route=app_route,
        message={"array_name": array_name, "job_name": job_name, "limit": limit},
        request_type="get",
    )

    return resp["array_tasks"]


def resume_workflow_from_id(
    workflow_id: int,
    cluster_name: str,
    reset_if_running: bool = True,
    log: bool = True,
    timeout: int = 180,
) -> None:
    """Given a workflow ID, resume the workflow.

    Raise an error if the workflow is not completed successfully on resume.
    """
    if log:
        JobmonLoggerConfig.attach_default_handler(
            logger_name="jobmon", log_level=logging.INFO
        )

    factory = WorkflowRunFactory(workflow_id=workflow_id)

    # Signal for a resume - move existing workflow runs to C or H resume depending on the input
    factory.set_workflow_resume(reset_running_jobs=reset_if_running)
    factory.reset_task_statuses(reset_if_running=reset_if_running)
    # Create the client workflow run
    new_wfr = factory.create_workflow_run()

    # Create swarm
    swarm = SwarmWorkflowRun(
        workflow_run_id=new_wfr.workflow_run_id, status=new_wfr.status
    )
    swarm.from_workflow_id(workflow_id)

    with DistributorContext(
        workflow_run_id=new_wfr.workflow_run_id,
        cluster_name=cluster_name,
        timeout=timeout,
    ) as distributor:
        swarm.run(distributor_alive_callable=distributor.alive)

    # Check on the swarm status - raise an error if != "D"
    if swarm.status == WorkflowRunStatus.DONE:
        print(f"Workflow {workflow_id} has successfully resumed to completion.")
    else:
        raise WorkflowRunStateError(
            f"Workflow run {swarm.workflow_run_id}, associated with workflow {workflow_id}",
            f"failed with status {swarm.status}",
        )



"""A framework that many tasks have in common while varying by a declared set of arguments."""

from __future__ import annotations

import copy
import hashlib
from http import HTTPStatus as StatusCodes
import logging
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TYPE_CHECKING,
    Union,
)

import yaml

from jobmon.client.array import Array
from jobmon.client.node import Node
from jobmon.client.task import Task, validate_task_resource_scales
from jobmon.client.task_template_version import TaskTemplateVersion
from jobmon.core.constants import ExecludeTTVs, MaxConcurrentlyRunning
from jobmon.core.exceptions import InvalidResponse
from jobmon.core.requester import Requester
from jobmon.core.serializers import (
    SerializeClientTaskTemplate,
    SerializeTaskTemplateResourceUsage,
)

if TYPE_CHECKING:
    from jobmon.client.tool_version import ToolVersion

logger = logging.getLogger(__name__)


class TaskTemplate:
    """Task Template outlines the structure of a Task to give it more context within the DAG.

    A Task Template defines a framework that many tasks have in common while varying by a
    declared set of arguments.
    """

    def __init__(
        self, template_name: str, requester: Optional[Requester] = None
    ) -> None:
        """Groups tasks of a type.

        Declares the concrete arguments that instances may vary over either from workflow to
        workflow or between nodes in the stage of a dag.

        Args:
            template_name: the name of this task template.
            requester: object to communicate with the flask services.
        """
        self.template_name = template_name
        self._task_template_id: int
        self._tool_version: ToolVersion

        # versions
        self._task_template_versions: List[TaskTemplateVersion] = []
        self._active_task_template_version: TaskTemplateVersion

        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

    @classmethod
    def get_task_template(
        cls: Type[TaskTemplate],
        tool_version: ToolVersion,
        template_name: str,
    ) -> TaskTemplate:
        """Get a bound instance of TaskTemplate.

        Args:
            tool_version: ToolVersion, to associate this task template with
            template_name: name of this specific task template
        """
        task_template = cls(template_name, tool_version.requester)
        task_template.bind(tool_version)
        return task_template

    @classmethod
    def from_wire(
        cls: Type[TaskTemplate], wire_tuple: Tuple, tool_version: ToolVersion
    ) -> TaskTemplate:
        """Get a bound instance of TaskTemplate from the http wire format.

        Args:
            wire_tuple: Wire format for ToolVersion defined in jobmon.serializers.
            tool_version: ToolVersion, to associate this task template with
            requester: communicate with the flask services.
        """
        task_template_kwargs = SerializeClientTaskTemplate.kwargs_from_wire(wire_tuple)

        if task_template_kwargs["tool_version_id"] != tool_version.id:
            raise ValueError(
                "tool_version_id in wire_tuple does not match tool_version object"
                f". Expected {tool_version.id} in wire_tuple. Got "
                f"{task_template_kwargs['tool_version_id']}"
            )

        task_template = cls(
            template_name=task_template_kwargs["template_name"],
            requester=tool_version.requester,
        )
        task_template._task_template_id = task_template_kwargs["id"]
        task_template._tool_version = tool_version
        return task_template

    def bind(self, tool_version: ToolVersion) -> None:
        """Bind task template to the db.

        Args:
            tool_version: the ToolVersion this task template is associated with.
        """
        if self.is_bound:
            return

        app_route = "/task_template"
        return_code, response = self.requester.send_request(
            app_route=app_route,
            message={
                "tool_version_id": tool_version.id,
                "task_template_name": self.template_name,
            },
            request_type="post",
        )

        if return_code != StatusCodes.OK:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST request through route "
                f"{app_route}. Expected code 200. Response content: {response}"
            )

        self._task_template_id = response["task_template_id"]
        self._tool_version = tool_version

    @property
    def is_bound(self) -> bool:
        """If the task template version has been bound to the database."""
        return hasattr(self, "_task_template_id")

    @property
    def id(self) -> int:
        """Unique id from db if task_template has been bound."""
        if not self.is_bound:
            raise AttributeError(
                "Cannot access id until TaskTemplate is bound to a ToolVersion"
            )
        return self._task_template_id

    @property
    def tool_version(self) -> ToolVersion:
        """The ToolVersion this task_template has been bound to."""
        if not self.is_bound:
            raise AttributeError(
                "Cannot access tool_version until TaskTemplate is bound to a ToolVersion"
            )
        return self._tool_version

    @property
    def task_template_versions(self) -> List[TaskTemplateVersion]:
        """Version of task template if it has been bound."""
        return self._task_template_versions

    @property
    def active_task_template_version(self) -> TaskTemplateVersion:
        """The TaskTemplateVersion to use when spawning tasks."""
        if not self.task_template_versions:
            raise AttributeError(
                "Cannot access attribute active_task_template_version because there are no "
                f"TaskTemplateVersions associated with task_template_name={self.template_name}"
                ". Either create some using get_task_template_version or load existing ones "
                "from the database using load_task_template_versions."
            )
        return self._active_task_template_version

    @property
    def default_cluster_name(self) -> str:
        """Default cluster_name associated with active tool version."""
        return self.active_task_template_version.default_cluster_name

    @default_cluster_name.setter
    def default_cluster_name(self, cluster_name: str) -> None:
        """Set default cluster.

        Args:
            cluster_name: name of cluster to set as default.
        """
        self.active_task_template_version.default_cluster_name = cluster_name

    @property
    def default_max_attempts(self) -> Optional[int]:
        """Default max attempts of the active tool version."""
        if self.active_task_template_version.default_max_attempts is None:
            if self.tool_version.default_max_attempt:
                self.active_task_template_version.set_default_max_attempts(
                    self.tool_version.default_max_attempt
                )
        return self.active_task_template_version.default_max_attempts

    def set_default_max_attempts(self, value: int) -> None:
        """Set default max_attempts.

        Args:
            value: value of max_attempts.
        """
        self.active_task_template_version.default_max_attempts = value

    @property
    def default_compute_resources_set(self) -> Dict[str, Dict[str, Any]]:
        """Default compute resources associated with active tool version."""
        return self.active_task_template_version.default_compute_resources_set

    @property
    def default_resource_scales_set(self) -> Dict[str, Dict[str, float]]:
        """Default resource scales associated with active tool version."""
        return self.active_task_template_version.default_resource_scales_set

    def update_default_compute_resources(
        self, cluster_name: str, **kwargs: Any
    ) -> None:
        """Update default compute resources in place only overridding specified keys.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to modify default values for.
            **kwargs: any key/value pair you want to update specified as an argument.
        """
        self.active_task_template_version.update_default_compute_resources(
            cluster_name, **kwargs
        )

    def update_default_resource_scales(self, cluster_name: str, **kwargs: Any) -> None:
        """Update default resource scales in place only overridding specified keys.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to modify default values for.
            **kwargs: any key/value pair you want to update specified as an argument.
        """
        self.active_task_template_version.update_default_resource_scales(
            cluster_name, **kwargs
        )

    def set_default_compute_resources_from_yaml(
        self, yaml_file: str, default_cluster_name: str = ""
    ) -> None:
        """Set default ComputeResources from a user provided yaml file for task template level.

        Args:
            default_cluster_name: name of cluster to set default values for.
            yaml_file: the yaml file that is providing the compute resource values.
        """
        if self.default_cluster_name is None and default_cluster_name is None:
            raise ValueError(
                "Must specify default_cluster_name when using default_compute_resources "
                "option. Set in tool.get_task_template() or "
                "set_default_compute_resources_from_yaml"
            )

        # Take passed-in default_cluster_name over task_template.default_cluster_name
        elif not default_cluster_name:
            default_cluster_name = self.default_cluster_name

        # Read in compute resources from YAML
        with open(yaml_file, "r") as stream:
            try:
                compute_resources = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                raise ValueError(
                    f"Unable to read compute resources from {yaml_file}."
                ) from exc

        self.active_task_template_version.set_default_compute_resources_from_dict(
            self.default_cluster_name,
            (
                compute_resources["task_template_resources"][self.template_name][
                    default_cluster_name
                ]
            ),
        )

    def set_default_resource_scales_from_yaml(
        self, yaml_file: str, default_cluster_name: str = ""
    ) -> None:
        """Set default Resource Scales from a user provided yaml file for task template level.

        Args:
            default_cluster_name: name of cluster to set default values for.
            yaml_file: the yaml file that is providing the compute resource values.
        """
        if self.default_cluster_name is None and default_cluster_name is None:
            raise ValueError(
                "Must specify default_cluster_name when using default_resource_scales "
                "option. Set in tool.get_task_template() or "
                "set_default_resource_scales_from_yaml"
            )

        # Take passed-in default_cluster_name over task_template.default_cluster_name
        elif not default_cluster_name:
            default_cluster_name = self.default_cluster_name

        # Read in resource scales from YAML
        with open(yaml_file, "r") as stream:
            try:
                resource_scales = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                raise ValueError(
                    f"Unable to read compute resources from {yaml_file}."
                ) from exc

        self.active_task_template_version.set_default_resource_scales_from_dict(
            self.default_cluster_name,
            (
                resource_scales["task_template_scales"][self.template_name][
                    default_cluster_name
                ]
            ),
        )

    def set_default_compute_resources_from_dict(
        self, cluster_name: str, compute_resources: Dict[str, Any]
    ) -> None:
        """Set default compute resources for a given cluster_name.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to set default values for.
            compute_resources: dictionary of default compute resources to run tasks
                with. Can be overridden at task level. dict of {resource_name: resource_value}
        """
        self.default_cluster_name = cluster_name
        self.active_task_template_version.set_default_compute_resources_from_dict(
            cluster_name, compute_resources
        )

    def set_default_resource_scales_from_dict(
        self, cluster_name: str, resource_scales: Dict[str, float]
    ) -> None:
        """Set default resource scales for a given cluster_name.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to set default values for.
            resource_scales: dictionary of default resource scales to adjust task
                resources with. Can be overridden at task level.
        """
        self.default_cluster_name = cluster_name
        self.active_task_template_version.set_default_resource_scales_from_dict(
            cluster_name, resource_scales
        )

    def set_active_task_template_version_id(
        self, task_template_version_id: Union[str, int] = "latest"
    ) -> None:
        """The TaskTemplateVersion that is set as the active one (latest is default).

        Args:
            task_template_version_id: which version to set as active on this object.
        """
        if not self.task_template_versions:
            raise AttributeError(
                "Cannot set an active_task_template_version until task templates versions "
                f"exist for task_template_name={self.template_name}. Either create some using "
                " get_task_template_version or load existing ones from the database using "
                "load_task_template_versions."
            )
        version_index_lookup = {
            self.task_template_versions[index].id: index
            for index in range(len(self.task_template_versions))
            if self.task_template_versions[index].is_bound
        }

        # get the lookup value
        if task_template_version_id == "latest":
            lookup_version: int = int(max(version_index_lookup.keys()))
        else:
            lookup_version = int(task_template_version_id)

        # check that the version exists
        try:
            version_index = version_index_lookup[lookup_version]
        except KeyError:
            raise ValueError(
                f"task_template_version_id {task_template_version_id} is not a valid "
                f"version for task_template_name={self.template_name} and tool_version_id="
                f"{self._tool_version.id}. Valid versions={version_index_lookup.keys()}"
            )

        self._active_task_template_version = self.task_template_versions[version_index]

    def set_active_task_template_version(
        self, task_template_version: TaskTemplateVersion
    ) -> None:
        """The TaskTemplateVersion that is set as the active one.

        Args:
            task_template_version: which version to set as active on this object.
        """
        # build a lookup between the version list and the version hash
        hash_index_lookup = {
            hash(self.task_template_versions[index]): index
            for index in range(len(self.task_template_versions))
        }

        # get the lookup value
        lookup_hash = hash(task_template_version)

        # check that the version exists
        try:
            version_index = hash_index_lookup[lookup_hash]
        except KeyError:
            self.task_template_versions.append(task_template_version)
            version_index = len(self.task_template_versions) - 1
            if version_index < 0:
                version_index = 0
            hash_index_lookup[lookup_hash] = version_index

        self._active_task_template_version = self.task_template_versions[version_index]

    def load_task_template_versions(self) -> None:
        """Load task template versions associated with this task template from the database."""
        app_route = f"/task_template/{self.id}/versions"
        return_code, response = self.requester.send_request(
            app_route=app_route, message={}, request_type="get"
        )

        if return_code != StatusCodes.OK:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST request through route "
                f"{app_route}. Expected code 200. Response content: {response}"
            )

        task_template_versions = [
            TaskTemplateVersion.from_wire(wire_args, self)
            for wire_args in response["task_template_versions"]
        ]
        self._task_template_versions = task_template_versions

        # activate the latest version
        if self.task_template_versions:
            self.set_active_task_template_version_id()

    def get_task_template_version(
        self,
        command_template: str,
        node_args: Optional[List[str]] = None,
        task_args: Optional[List[str]] = None,
        op_args: Optional[List[str]] = None,
        default_cluster_name: str = "",
        default_compute_resources: Optional[Dict[str, Any]] = None,
        default_resource_scales: Optional[Dict[str, float]] = None,
        default_max_attempts: Optional[int] = None,
    ) -> TaskTemplateVersion:
        """Create a task template version instance. If it already exists, activate it.

        Args:
            command_template: an abstract command representing a task, where the arguments to
                the command have defined names but the values are not assigned.
                eg: '{python} {script} --data {data} --para {para} {verbose}'
            node_args: any named arguments in command_template that make the command unique
                within this template for a given workflow run. Generally these are arguments
                that can be parallelized over.
            task_args: any named arguments in command_template that make the command unique
                across workflows if the node args are the same as a previous workflow.
                Generally these are arguments about data moving though the task.
            op_args: any named arguments in command_template that can change without changing
                the identity of the task. Generally these are things like the task executable
                location or the verbosity of the script.
            default_cluster_name: the default cluster to run each task associated with this
                template on.
            default_compute_resources: dictionary of default compute resources to run tasks
                with. Can be overridden at task level. dict of {resource_name: resource_value}.
                Must specify default_cluster_name when this option is used.
            default_resource_scales: dictionary of default resource scales to adjust task
                resources with. Can be overridden at task level.
                dict of {resource_name: scale_factor}. Scale factor can be a numeric value, a
                Callable that will be applied to the existing resources, or an Iterator. Any
                Callable should take a single numeric value as its sole argument. Any
                Iterator should only yield numeric values. Any Iterable can be easily
                converted to an Iterator by using the iter() built-in (e.g. iter([80, 160,
                190])).
            default_max_attempts: default max_attempts associated with this template on.
        """
        if default_compute_resources is not None and not default_cluster_name:
            raise ValueError(
                "Must specify default_cluster_name when using default_compute_resources option"
            )

        if not node_args:
            node_args = []
        if not task_args:
            task_args = []
        if not op_args:
            op_args = []

        task_template_version = TaskTemplateVersion.get_task_template_version(
            self,
            command_template=command_template,
            node_args=node_args,
            task_args=task_args,
            op_args=op_args,
        )
        # set compute resources on the task template version if specified
        task_template_version.default_cluster_name = default_cluster_name
        task_template_version.default_max_attempts = default_max_attempts
        if default_compute_resources:
            task_template_version.set_default_compute_resources_from_dict(
                default_cluster_name, default_compute_resources
            )
        if default_resource_scales:
            task_template_version.set_default_resource_scales_from_dict(
                default_cluster_name, default_resource_scales
            )

        # now activate it
        self.set_active_task_template_version(task_template_version)

        return self.active_task_template_version

    def create_task(
        self,
        name: str = "",
        upstream_tasks: List[Task] = [],
        task_attributes: Union[List, dict] = {},
        max_attempts: Optional[int] = None,
        compute_resources: Optional[Dict[str, Any]] = None,
        compute_resources_callable: Optional[Callable] = None,
        resource_scales: Optional[Dict[str, Any]] = None,
        cluster_name: str = "",
        fallback_queues: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Task:
        """Create an instance of a task associated with this template.

        Args:
            name: a name associated with this specific task
            upstream_tasks: Task objects that must be run prior to this one
            task_attributes (dict or list): attributes and their values or just the attributes
                that will be given values later
            max_attempts: Number of attempts to try this task before giving up. Default is 3.
            cluster_name: name of cluster to run task on.
            compute_resources: dictionary of default compute resources to run tasks
                with. Can be overridden at task template or task level.
                dict of {resource_name: resource_value}
            compute_resources_callable: compute resources generating callable.
            resource_scales (dict): How much users want to scale their resource request if the
                the initial request fails. Scale factor can be a numeric value, a Callable
                that will be applied to the existing resources, or an Iterator. Any Callable
                should take a single numeric value as its sole argument. Any Iterator should
                only yield numeric values. Any Iterable can be easily converted to an
                Iterator by using the iter() built-in (e.g. iter([80, 160, 190])).

            **kwargs: values for each argument specified in command_template

        Returns:
            ExecutableTask

        Raises:
            ValueError: if the args that are supplied do not match the args in the command
                template.
        """
        # make sure task template is bound to tool version
        if not self.is_bound:
            raise RuntimeError(
                f"TaskTemplate={self.template_name} must be bound to a tool version before "
                "tasks can be created."
            )

        # bind task template version to task template if needed
        if not self.active_task_template_version.is_bound:
            self.active_task_template_version.bind(self)

        # kwargs quality assurance
        if self.active_task_template_version.template_args != set(kwargs.keys()):
            raise ValueError(
                f"Unexpected kwarg. Task Template requires "
                f"{self.active_task_template_version.template_args}, got {set(kwargs.keys())}"
            )

        # validate non-empty resource_scale dicts
        if resource_scales:
            validate_task_resource_scales(resource_scales=resource_scales)

        node_args = self.active_task_template_version.filter_kwargs(
            "node_args", **kwargs
        )
        task_args = self.active_task_template_version.filter_kwargs(
            "task_args", **kwargs
        )
        op_args = self.active_task_template_version.filter_kwargs("op_args", **kwargs)

        # build node
        node = Node(self.active_task_template_version, node_args, self.requester)

        # build task
        task = Task(
            node=node,
            task_args=task_args,
            op_args=op_args,
            compute_resources=compute_resources,
            compute_resources_callable=compute_resources_callable,
            resource_scales=copy.deepcopy(resource_scales),
            cluster_name=cluster_name,
            name=name,
            max_attempts=max_attempts,
            upstream_tasks=upstream_tasks,
            task_attributes=task_attributes,
            requester=self.requester,
            fallback_queues=fallback_queues,
        )
        return task

    def create_tasks(
        self,
        max_attempts: Optional[int] = None,
        upstream_tasks: Optional[List[Task]] = None,
        max_concurrently_running: int = MaxConcurrentlyRunning.MAXCONCURRENTLYRUNNING,
        compute_resources: Optional[Dict[str, Any]] = None,
        compute_resources_callable: Optional[Callable] = None,
        resource_scales: Optional[Dict[str, Any]] = None,
        cluster_name: str = "",
        name: Optional[str] = None,
        **kwargs: Any,
    ) -> List[Task]:
        """Creates a set of tasks equal to the cross product of all node args.

        Args:
            max_attempts: the max number of attempts a task in the array can be retried
            upstream_tasks: dependencies for all tasks in this array
            max_concurrently_running: the max number of tasks that can run at once
            compute_resources: resources to associate with this array, if different from
                the task template default resources
            compute_resources_callable: a function that can dynamically generate resources on
                retries, if different from the task template callable.
            resource_scales: Parameters for how aggressive we will rescale tasks that
                are resource killed.
            cluster_name: The cluster the array will run on
            **kwargs: task, node, and op_args as defined in the command template. If you
                provide node_args as an iterable, they will be expanded.
            name: the name of the array.
        """
        if upstream_tasks is None:
            upstream_tasks = []

        # kwargs quality assurance
        if not set(kwargs.keys()).issuperset(
            self.active_task_template_version.task_args
        ):
            raise ValueError(
                f"Missing task_args for this array. Task Template requires task_args="
                f"{self.active_task_template_version.task_args}, got {set(kwargs.keys())}."
            )
        if not set(kwargs.keys()).issuperset(self.active_task_template_version.op_args):
            raise ValueError(
                f"Missing op_args for this array. Task Template requires op_args="
                f"{self.active_task_template_version.op_args}, got {set(kwargs.keys())}."
            )

        # validate non-empty resource_scale dicts
        if resource_scales:
            validate_task_resource_scales(resource_scales=resource_scales)

        # Split node, task, and op_args
        node_args = self.active_task_template_version.filter_kwargs(
            "node_args", **kwargs
        )
        task_args = self.active_task_template_version.filter_kwargs(
            "task_args", **kwargs
        )
        op_args = self.active_task_template_version.filter_kwargs("op_args", **kwargs)

        array = Array(
            task_template_version=self.active_task_template_version,
            task_args=task_args,
            op_args=op_args,
            cluster_name=cluster_name,
            max_concurrently_running=max_concurrently_running,
            upstream_tasks=upstream_tasks,
            compute_resources=compute_resources,
            compute_resources_callable=compute_resources_callable,
            resource_scales=resource_scales,
            name=name,
        )

        # Create tasks on the array
        tasks = array.create_tasks(
            upstream_tasks=upstream_tasks,
            max_attempts=max_attempts,
            resource_scales=resource_scales,
            **node_args,
        )
        return tasks

    def __hash__(self) -> int:
        """A hash of the TaskTemplate name and tool version concatenated together."""
        hash_value = int(
            hashlib.sha256(self.template_name.encode("utf-8")).hexdigest(), 16
        )
        return hash_value

    def resource_usage(
        self,
        workflows: Optional[List[int]] = None,
        node_args: Optional[Dict[str, Any]] = None,
        ci: Optional[float] = None,
    ) -> Optional[dict]:
        """Get the aggregate resource usage for a TaskTemplate."""
        message: Dict[Any, Any] = dict()
        message["task_template_version_id"] = self._active_task_template_version.id

        # exclude ttv with huge number of tasks
        exclue_list = ExecludeTTVs.EXECLUDE_TTVS
        if self._active_task_template_version.id in exclue_list:
            msg = (
                f"Resource usage query for task_template_version "
                f"{self._active_task_template_version.id}"
                f"  is restricted."
            )
            logger.warning(msg)
            return None

        if workflows:
            message["workflows"] = workflows
        if node_args:
            message["node_args"] = node_args
        if ci:
            message["ci"] = ci
        app_route = "/task_template_resource_usage"
        return_code, response = self.requester.send_request(
            app_route=app_route, message=message, request_type="post"
        )
        if return_code != StatusCodes.OK:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from GET "
                f"request through route {app_route}. Expected code "
                f"200. Response content: {response}"
            )

        def format_bytes(value: Any) -> Optional[str]:
            if value is not None:
                return str(value) + "B"
            else:
                return value

        kwargs = SerializeTaskTemplateResourceUsage.kwargs_from_wire(response)
        resources = {
            "num_tasks": kwargs["num_tasks"],
            "min_mem": format_bytes(kwargs["min_mem"]),
            "max_mem": format_bytes(kwargs["max_mem"]),
            "mean_mem": format_bytes(kwargs["mean_mem"]),
            "min_runtime": kwargs["min_runtime"],
            "max_runtime": kwargs["max_runtime"],
            "mean_runtime": kwargs["mean_runtime"],
            "median_mem": format_bytes(kwargs["median_mem"]),
            "median_runtime": kwargs["median_runtime"],
            "ci_mem": kwargs["ci_mem"],
            "ci_runtime": kwargs["ci_runtime"],
        }
        return resources

    def __repr__(self) -> str:
        """A representation string for a TaskTemplate instance."""
        repr_string = (
            f"TaskTemplate(" f"task_template_versions={self.task_template_versions}"
        )

        try:
            repr_string += f", tool_version_id={self.tool_version.id}"
            repr_string += f", id={self.id})"
        except AttributeError:
            # Bind somehow not called, so terminate the repr string
            repr_string += ")"
        return repr_string



import os
from typing import Any, Dict, Optional

from jobmon.core.requester import Requester


def add_task_attributes(
    attr_dict: Dict[str, Any], task_id: Optional[int] = None
) -> None:
    """Add task attributes to database.

    Args:
        attr_dict: Key/value pairs to store as task attributes. must be coercable to json.
        task_id: The task id to associate the attributes to. If not provided, will check the
            shell environment for a task id. Otherwise will raise a ValueError

    Raises: ValueError, TypeError
    """
    try:
        if task_id is None:
            task_id = int(os.environ["JOBMON_TASK_ID"])
    except KeyError:
        raise ValueError(
            "JOBMON_TASK_ID not found in environment. Must specify task_id if not calling in "
            "the context of a task."
        )

    # add task id to the attribute dict
    attr_dict["task_id"] = task_id

    requester = Requester.from_defaults()
    requester.send_request(
        app_route="/task/bind_task_attributes", message=attr_dict, request_type="post"
    )



"""The client Task Resources with the resources initiation and binding to Task ID."""

from __future__ import annotations

import hashlib
from http import HTTPStatus as StatusCodes
import json
import logging
from math import ceil
import numbers
import re
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

from jobmon.client.units import MemUnit, TimeUnit
from jobmon.core.cluster_protocol import ClusterQueue
from jobmon.core.exceptions import InvalidResponse
from jobmon.core.requester import Requester


logger = logging.getLogger(__name__)


class TaskResources:
    """An object representing the resources for a specific task."""

    def __init__(
        self,
        requested_resources: Dict[str, Any],
        queue: ClusterQueue,
        requester: Optional[Requester] = None,
    ) -> None:
        """Initialize the task resource object."""
        for resource, value in requested_resources.items():
            if resource == "memory":
                requested_resources[resource] = self.convert_memory_to_gib(value)
            if resource == "runtime":
                requested_resources[resource] = self.convert_runtime_to_s(value)
        self.requested_resources = requested_resources
        self.queue = queue

        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

    @property
    def is_bound(self) -> bool:
        """If the TaskResources has been bound to the database."""
        return hasattr(self, "_id")

    @property
    def id(self) -> int:
        """If the task resources has been bound to the database."""
        if not self.is_bound:
            raise AttributeError(
                "Cannot access id until TaskResources is bound to database"
            )
        return self._id

    def bind(self) -> None:
        """Bind TaskResources to the database."""
        # Check if it's already been bound
        if self.is_bound:
            logger.debug(
                "This task resource has already been bound, and assigned"
                f"task_resources_id {self.id}"
            )
            return

        app_route = "/task/bind_resources"
        msg = {
            "queue_id": self.queue.queue_id,
            "task_resources_type_id": "O",
            "requested_resources": self.requested_resources,
        }
        return_code, response = self.requester.send_request(
            app_route=app_route, message=msg, request_type="post"
        )

        if return_code != StatusCodes.OK:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST "
                f"request through route {app_route}. Expected "
                f"code 200. Response content: {response}"
            )
        self._id = response

    def validate_resources(
        self: TaskResources, strict: bool = False
    ) -> Tuple[bool, str]:
        is_valid, msg = self.queue.validate_resources(
            strict, **self.requested_resources
        )
        return is_valid, msg

    def coerce_resources(self: TaskResources) -> TaskResources:
        """Coerce TaskResources to fit on queue. If resources change return a new object."""
        valid_resources = self.queue.coerce_resources(**self.requested_resources)
        coerced_task_resources = self.__class__(valid_resources, self.queue)
        if coerced_task_resources != self:
            return coerced_task_resources
        else:
            return self

    def adjust_resources(
        self: TaskResources,
        resource_scales: Dict[
            str, Union[numbers.Number, Callable, Iterator[numbers.Number]]
        ],
        fallback_queues: Optional[List[ClusterQueue]] = None,
    ) -> TaskResources:
        """Adjust TaskResources after a resource error, returning a new object if it changed.

        Args:
            resource_scales: Specifies how much to scale the failed Task's resources by.
                Scale factor can be a numeric value, a Callable that will be applied
                to the existing resources, or an Iterator. Any Callable should take
                a single numeric value as its sole argument. Any Iterator should only yield
                numeric values. Any Iterable can be easily converted to an Iterator by using
                the iter() built-in (e.g. iter([80, 160, 190])).
            fallback_queues: list of queues that users specify. If their jobs exceed the
                resources of a given queue, Jobmon will try to run their jobs on the fallback
                queues.
        """
        if fallback_queues is None:
            fallback_queues = []
        existing_resources = self.requested_resources.copy()
        resource_updates: Dict[str, Any] = {}

        # Only cores, memory, and runtime get scaled
        for resource, scaler in resource_scales.items():
            if resource in existing_resources.keys():
                if isinstance(scaler, numbers.Number):
                    new_resource_value = self.scale_val(
                        existing_resources[resource], scaler  # type: ignore
                    )
                elif callable(scaler):
                    new_resource_value = scaler(existing_resources[resource])
                elif isinstance(scaler, Iterator):
                    try:
                        new_resource_value = next(scaler)  # type: ignore
                    except StopIteration:
                        logger.warning(
                            "Not enough elements left in Iterator, re-using previous value "
                            f"for {resource}: {existing_resources[resource]}"
                        )
                        new_resource_value = existing_resources[resource]
                else:
                    raise ValueError(
                        "Keys in the resource_scales dictionary must be either numeric "
                        f"values, Iterators, or Python Callables; found {scaler}, type "
                        f"{type(scaler)} instead."
                    )
                if not isinstance(new_resource_value, numbers.Number):
                    raise ValueError(
                        "Attemping to update resource to a non-numeric value, "
                        f"{new_resource_value}. If passing an Iterator, elements must be "
                        "numeric values. If passing a Callable, return value must be a "
                        "numeric value."
                    )
                resource_updates[resource] = new_resource_value

        scaled_resources = dict(existing_resources, **resource_updates)

        # If it fails, try the fallback queues.
        queues = [self.queue] + fallback_queues
        while queues:
            next_queue = queues.pop(0)
            is_valid, _ = next_queue.validate_resources(strict=True, **scaled_resources)
            if is_valid:
                valid_resources = scaled_resources
                break
        else:  # no break
            # We've run out of queues so use the final queue and coerce
            valid_resources = next_queue.coerce_resources(**scaled_resources)

        adjust_resources = self.__class__(valid_resources, next_queue)
        if adjust_resources != self:
            return adjust_resources
        else:
            return self

    @staticmethod
    def convert_memory_to_gib(memory_str: str) -> int:
        """Given a memory request with a unit suffix, convert to GiB."""
        try:
            # User could pass in a raw value for memory, assume to be in GiB.
            # This is also the path taken by adjust
            return int(memory_str)
        except ValueError:
            return MemUnit.convert(memory_str, to="G")

    @staticmethod
    def convert_runtime_to_s(time_str: Union[str, float, int]) -> int:
        """Given a runtime request, coerce to seconds for recording in the DB."""
        try:
            # If a numeric is provided, assumed to be in seconds
            return int(time_str)
        except ValueError:
            time_str = str(time_str).lower()

            # convert to seconds if its datetime with a supported format
            try:
                hours, minutes, seconds = time_str.split(":")
                time_seconds = (
                    TimeUnit.hour_to_sec(int(hours))
                    + TimeUnit.min_to_sec(int(minutes))
                    + int(seconds)
                )
                return time_seconds
            except ValueError:
                try:
                    raw_value, unit = re.findall(r"[A-Za-z]+|\d+", time_str)
                except ValueError:
                    # Raised if there are not exactly 2 values to unpack from above regex
                    raise ValueError(
                        "The provided runtime request must be in a format of numbers "
                        "followed by one or two characters indicating the unit. "
                        "E.g. 1h, 60m, 3600s."
                    )

                if "h" in unit:
                    # Hours provided
                    return TimeUnit.hour_to_sec(int(raw_value))
                elif "m" in unit:
                    # Minutes provided
                    return TimeUnit.min_to_sec(int(raw_value))
                elif "s" in unit:
                    return int(raw_value)
                else:
                    raise ValueError("Expected one of h, m, s as the suffixed unit.")

    @staticmethod
    def scale_val(val: int, scaling_factor: float) -> float:
        """Used ceil instead of round or floor, to handle case when resources is 1.

        For example, if runtime was 1, resource scales was 0.2. Then the resource would adjust
        to 1.2, which would be truncated to 1 again if using floor/round.
        """
        return int(ceil(val * (1 + scaling_factor)))

    def __hash__(self) -> int:
        """Determine the hash of a task resources object."""
        # Note: this algorithm assumes all keys and values in the resources dict are
        # JSON-serializable. Since that's a requirement for logging in the database,
        # this assumption should be safe.

        # Uniqueness is determined by queue name and the resources parameter.
        if not hasattr(self, "_hash_val"):
            hashval = hashlib.sha256()
            hashval.update(bytes(str(hash(self.queue.queue_name)).encode("utf-8")))
            resources_str = str(
                hash(json.dumps(self.requested_resources, sort_keys=True))
            )
            hashval.update(bytes(resources_str.encode("utf-8")))
            self._hash_val = int(hashval.hexdigest(), 16)
        return self._hash_val

    def __eq__(self, other: object) -> bool:
        """Check equality of task resources objects."""
        if not isinstance(other, TaskResources):
            return False
        return hash(self) == hash(other)

    def __repr__(self) -> str:
        """A representation string for a TaskResources instance."""
        repr_string = (
            f"TaskResources(queue={self.queue.queue_name}, "
            f"requested_resources={self.requested_resources}"
        )

        try:
            repr_string += f", id={self.id})"
        except AttributeError:
            repr_string += ")"
        return repr_string



"""A node represents an individual task within a DAG."""

from __future__ import annotations

import hashlib
import logging
from typing import Any, Dict, List, Optional, Set

from jobmon.client.task_template_version import TaskTemplateVersion
from jobmon.core.constants import SpecialChars
from jobmon.core.requester import Requester


logger = logging.getLogger(__name__)


class Node:
    """A node represents an individual task within a Dag."""

    def __init__(
        self,
        task_template_version: TaskTemplateVersion,
        node_args: Dict[str, Any],
        requester: Optional[Requester] = None,
    ) -> None:
        """A node represents an individual task within a Dag.

        This includes its relationship to other nodes that it is dependent upon or nodes that
        depend upon it. A node stores node arguments (arguments relating to the
        actual shape of the dag and number of tasks created for a given
        stage/task template) and can register itself with the database via the
        Jobmon Query Service and the Jobmon State Manager.

        Args:
            task_template_version: The associated TaskTemplateVersion.
            node_args: key-value pairs of arg_name and a value.
            requester: Requester object to communicate with the flask services.
        """
        self._node_id: Optional[int] = None
        self.task_template_version = task_template_version
        self.node_args = node_args
        self.mapped_node_args = self.task_template_version.convert_arg_names_to_ids(
            **self.node_args
        )
        self.node_args_hash = self._hash_node_args()
        self.upstream_nodes: Set[Node] = set()
        self.downstream_nodes: Set[Node] = set()

        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

    @property
    def node_id(self) -> int:
        """Unique id for each node."""
        if self._node_id is None:
            raise AttributeError("node_id cannot be accessed before node is bound")
        return self._node_id

    @node_id.setter
    def node_id(self, val: int) -> None:
        """Unique id for each node."""
        self._node_id = val

    @property
    def task_template_version_id(self) -> int:
        return self.task_template_version.id

    @property
    def default_name(self) -> str:
        """The default name of this node in the array."""
        name = (
            self.task_template_version.task_template.template_name
            + "_"
            + "_".join(
                [str(k) + "-" + str(self.node_args[k]) for k in self.node_args.keys()]
            )
        )

        # special char protection
        name = "".join(
            [
                c if c not in SpecialChars.ILLEGAL_SPECIAL_CHARACTERS else "_"
                for c in name
            ]
        )

        # long name protection
        name = name if len(name) < 250 else name[0:249]
        return name

    def _hash_node_args(self) -> int:
        """A hash of the node.

        The hash is the encoded result of the args and values concatenated together.
        """
        arg_ids = list(self.mapped_node_args.keys())
        arg_ids.sort()

        arg_values = [str(self.mapped_node_args[key]) for key in arg_ids]
        str_arg_ids = [str(arg) for arg in arg_ids]

        hash_value = int(
            hashlib.sha256(
                "".join(str_arg_ids + arg_values).encode("utf-8")
            ).hexdigest(),
            16,
        )
        return hash_value

    def add_upstream_node(self, upstream_node: Node) -> None:
        """Add a single node to this one's upstream Nodes.

        Args:
            upstream_node: node to add a dependency on
        """
        self.upstream_nodes.add(upstream_node)
        # Add this node to the upstream nodes' downstream
        upstream_node.downstream_nodes.add(self)

    def add_upstream_nodes(self, upstream_nodes: List[Node]) -> None:
        """Add many nodes to this one's upstream Nodes.

        Args:
            upstream_nodes: list of nodes to add dependencies on
        """
        for node in upstream_nodes:
            self.add_upstream_node(node)

    def add_downstream_node(self, downstream_node: Node) -> None:
        """Add a node to this one's downstream Nodes.

        Args:
            downstream_node: Node that will be dependent on this node
        """
        self.downstream_nodes.add(downstream_node)
        # avoid endless recursion, set directly
        downstream_node.upstream_nodes.add(self)

    def add_downstream_nodes(self, downstream_nodes: List[Node]) -> None:
        """Add a list of nodes as this node's downstream nodes.

        Args:
            downstream_nodes: Nodes that will be dependent on this node.
        """
        for node in downstream_nodes:
            self.add_downstream_node(node)

    def __repr__(self) -> str:
        """Repr the node attributes."""
        return (
            "Node("
            f"task_template_version_id={self.task_template_version_id}, "
            f"node_args={self.node_args}, "
            f"node_args_hash={self.node_args_hash})"
        )

    def __eq__(self, other: object) -> bool:
        """Check if two nodes have equal hashes."""
        if not isinstance(other, Node):
            return False
        else:
            return hash(self) == hash(other)

    def __lt__(self, other: Node) -> bool:
        """Check if this hash is less than anothers."""
        return hash(self) < hash(other)

    def __hash__(self) -> int:
        """Create a hash that will be a unique identifier for the node."""
        if not hasattr(self, "_hash_val"):
            hash_value = hashlib.sha256()
            hash_value.update(bytes(str(self.node_args_hash).encode("utf-8")))
            hash_value.update(bytes(str(self.task_template_version_id).encode("utf-8")))
            self._hash_val = int(hash_value.hexdigest(), 16)
        return self._hash_val



"""Swarm side task object."""

from __future__ import annotations

import logging
from typing import Callable, Dict, List, Optional, Set

from jobmon.client.task_resources import TaskResources
from jobmon.core.cluster import Cluster
from jobmon.core.cluster_protocol import ClusterQueue


logger = logging.getLogger(__name__)


class SwarmTask(object):
    """Swarm side task object."""

    def __init__(
        self,
        task_id: int,
        array_id: int,
        status: str,
        max_attempts: int,
        task_resources: TaskResources,
        cluster: Cluster,
        resource_scales: Optional[Dict] = None,
        fallback_queues: Optional[List[ClusterQueue]] = None,
        compute_resources_callable: Optional[Callable] = None,
    ) -> None:
        """Implementing swarm behavior of tasks.

        Args:
            task_id: id of task object from db auto increment.
            array_id: id of associated array object.
            status: status of task object.
            max_attempts: maximum number of task_instances before failure.
            task_resources: callable to be executed when Task is ready to be run and
                resources can be assigned.
            cluster: The name of the cluster that the user wants to run their tasks on.
            resource_scales: Specifies how much a user wants to scale their requested
                resources after failure.
            fallback_queues: A list of queues that users want to try if their original queue
                isn't able to handle their adjusted resources.
            compute_resources_callable: callable compute resources.
        """
        self.task_id = task_id
        self.array_id = array_id
        self.status = status

        self.downstream_swarm_tasks: Set[SwarmTask] = set()

        self.current_task_resources = task_resources
        self.compute_resources_callable = compute_resources_callable
        self.fallback_queues = fallback_queues
        self.resource_scales = resource_scales if resource_scales is not None else {}
        self.cluster = cluster

        self.max_attempts = max_attempts
        self.num_upstreams: int = 0
        self.num_upstreams_done: int = 0

    @property
    def all_upstreams_done(self) -> bool:
        """Return a bool of if upstreams are done or not."""
        if self.num_upstreams_done == self.num_upstreams:
            return True
        elif self.num_upstreams_done > self.num_upstreams:
            raise RuntimeError(
                "Error in dependency management. More upstream tasks done than exist in DAG."
            )
        else:
            return False

    @property
    def downstream_tasks(self) -> List[SwarmTask]:
        """Return list of downstream tasks."""
        return list(self.downstream_swarm_tasks)

    def __hash__(self) -> int:
        """Returns the ID of the task."""
        return self.task_id

    def __eq__(self, other: object) -> bool:
        """Check if the hashes of two tasks are equivalent."""
        if not isinstance(other, SwarmTask):
            return False
        else:
            return hash(self) == hash(other)

    def __lt__(self, other: SwarmTask) -> bool:
        """Check if one hash is less than the has of another Task."""
        return hash(self) < hash(other)



from __future__ import annotations

from collections.abc import Iterable
import copy
from http import HTTPStatus as StatusCodes
from itertools import product
import logging
from typing import Any, Callable, Dict, Iterator, List, Optional, TYPE_CHECKING, Union

from jobmon.client.node import Node
from jobmon.client.task import Task, validate_task_resource_scales
from jobmon.client.task_template_version import TaskTemplateVersion
from jobmon.core.constants import MaxConcurrentlyRunning
from jobmon.core.exceptions import InvalidResponse
from jobmon.core.requester import Requester


if TYPE_CHECKING:
    from jobmon.client.workflow import Workflow

logger = logging.getLogger(__name__)


class Array:
    """Representation of a client array object.

    Supports functionality to create tasks from the cross product of provided node_args.
    """

    compute_resources_callable: Callable[..., Any] | None

    def __init__(
        self,
        task_template_version: TaskTemplateVersion,
        task_args: Dict[str, Any],
        op_args: Dict[str, Any],
        cluster_name: str,
        max_concurrently_running: int = MaxConcurrentlyRunning.MAXCONCURRENTLYRUNNING,
        upstream_tasks: Optional[List[Task]] = None,
        compute_resources: Optional[Dict[str, Any]] = None,
        compute_resources_callable: Optional[Callable] = None,
        resource_scales: Optional[Dict[str, float]] = None,
        name: Optional[str] = None,
        requester: Optional[Requester] = None,
        max_attempts: Optional[int] = None,
    ) -> None:
        """Initialize the array object."""
        # task template attributes
        self.task_template_version = task_template_version

        # array name
        if name:
            self._name = name
        else:
            self._name = task_template_version.task_template.template_name
            self._name = self._name if len(self._name) < 255 else self._name[0:254]

        # array attributes
        self.max_concurrently_running = max_concurrently_running

        # task passthrough attributes
        self.task_args = task_args
        self.op_args = op_args

        # global upstreams
        if upstream_tasks is None:
            upstream_tasks = []
        self.upstream_tasks = upstream_tasks

        # compute resources
        if not cluster_name:
            cluster_name = self.task_template_version.default_cluster_name
        self._instance_cluster_name = cluster_name

        # max_attempts
        self._instance_max_attempts = max_attempts

        self._instance_compute_resource = (
            compute_resources if compute_resources is not None else {}
        )
        self.compute_resources_callable = compute_resources_callable
        self._instance_resource_scales = (
            resource_scales if resource_scales is not None else {}
        )

        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

        self.tasks: Dict[int, Task] = {}  # Initialize to empty dict

    @property
    def is_bound(self) -> bool:
        """If the array has been bound to the db."""
        return hasattr(self, "_array_id")

    @property
    def array_id(self) -> int:
        """If the array is bound then it will have been given an id."""
        if not self.is_bound:
            raise AttributeError("array_id cannot be accessed before workflow is bound")
        return self._array_id

    @array_id.setter
    def array_id(self, val: int) -> None:
        """Set the array id."""
        self._array_id = val

    @property
    def name(self) -> str:
        """Return the array name."""
        return self._name

    @property
    def compute_resources(self) -> Dict[str, Any]:
        """A dictionary that includes the users requested resources for the current run.

        E.g. {cores: 1, mem: 1, runtime: 60, queue: all.q}.
        """
        try:
            resources = self.workflow.default_compute_resources_set.get(
                self.cluster_name, {}
            ).copy()
        except AttributeError:
            resources = {}
        resources.update(
            self.task_template_version.default_compute_resources_set.get(
                self.cluster_name, {}
            ).copy()
        )
        resources.update(self._instance_compute_resource.copy())
        return resources

    @property
    def resource_scales(self) -> Dict[str, float]:
        """A dictionary that includes the users requested resource scales for the current run.

        E.g. {memory: 0.6, runtime: 0.3}.
        """
        try:
            scales = self.workflow.default_resource_scales_set.get(
                self.cluster_name, {}
            ).copy()
        except AttributeError:
            scales = {}
        scales.update(
            self.task_template_version.default_resource_scales_set.get(
                self.cluster_name, {}
            ).copy()
        )
        scales.update(self._instance_resource_scales.copy())
        return scales

    @property
    def cluster_name(self) -> str:
        """The name of the cluster the user wants to run their task on."""
        cluster_name = self._instance_cluster_name
        if not cluster_name:
            try:
                cluster_name = self.workflow.default_cluster_name
            except AttributeError:
                raise ValueError(
                    "cluster_name must be specified on workflow, task_template, or array"
                )
        return cluster_name

    @property
    def max_attempts(self) -> Optional[int]:
        """Get the max_attempts."""
        ma = self._instance_max_attempts
        if not ma:
            try:
                ma = self.workflow.default_max_attempts
            except AttributeError:
                raise ValueError(
                    "max_attempts must be specified on workflow, task_template, or array"
                )
        return ma

    @property
    def workflow(self) -> Workflow:
        """Get the workflow id if it has been bound to the db."""
        if not hasattr(self, "_workflow"):
            raise AttributeError(
                "workflow cannot be accessed via task before workflow is added to workflow"
            )
        return self._workflow

    @workflow.setter
    def workflow(self, val: Workflow) -> None:
        """Set the workflow id."""
        self._workflow = val

    def add_task(self, task: Task) -> None:
        """Add a task to an array.

        Set semantics - add tasks once only, based on hash name.

        Args:
            task: single task to add.
        """
        if task.cluster_name and self.cluster_name != task.cluster_name:
            raise ValueError(
                "Task assigned to different cluster than associated array. Task.cluster_name="
                f"{task.cluster_name}. Array.cluster_name={self.cluster_name}"
            )

        task_hash = hash(task)
        if task_hash in self.tasks.keys():
            raise ValueError(
                f"A task with hash {task_hash} already exists. All tasks in an Array must have"
                f" unique commands. Your command was: {task.command}"
            )

        self.tasks[task_hash] = task

        # populate backref
        task.array = self

    def create_tasks(
        self,
        upstream_tasks: Optional[List[Task]] = None,
        task_attributes: Union[List, dict] = {},
        max_attempts: Optional[int] = None,
        resource_scales: Optional[Dict[str, Any]] = None,
        **node_kwargs: Any,
    ) -> List[Task]:
        """Create a task associated with the array.

        Args:
            upstream_tasks: Task objects that must be run prior to this one
            task_attributes (dict or list): attributes and their values or just the attributes
                that will be given values later
            max_attempts: Number of attempts to try this task before giving up.
                Default is wf default.
            resource_scales: determines the scaling factor for how aggressive resource
                adjustments will be scaled up
            **node_kwargs: values for each node argument specified in command_template

        Raises:
            ValueError: if the args that are supplied do not match the args in the command
                template.
        """
        if upstream_tasks is None:
            # If not specified, defined from the array upstreams
            upstream_tasks = self.upstream_tasks

        # validate non-empty resource_scale dicts
        if resource_scales:
            validate_task_resource_scales(resource_scales=resource_scales)

        # Expand the node_args
        if not set(node_kwargs.keys()).issuperset(self.task_template_version.node_args):
            raise ValueError(
                f"Missing node_args for this array. Task Template requires node_args="
                f"{self.task_template_version.node_args}, got {set(node_kwargs.keys())}"
            )
        node_args_expanded = Array.expand_dict(**node_kwargs)

        # build tasks over node_args
        tasks = []
        for node_args in node_args_expanded:
            # build node
            node = Node(self.task_template_version, node_args, self.requester)

            task = Task(
                node=node,
                task_args=self.task_args,
                op_args=self.op_args,
                resource_scales=copy.deepcopy(resource_scales),
                max_attempts=max_attempts,
                upstream_tasks=upstream_tasks,
                task_attributes=task_attributes,
                requester=self.requester,
            )
            tasks.append(task)

            logger.debug(f"Adding Task {task}")
            if hash(task) in self.tasks.keys():
                raise ValueError(
                    f"A task with hash {hash(task)} already exists. "
                    f"All tasks in a workflow must have unique "
                    f"commands. Your command was: {task.command}"
                )
            self.add_task(task)

        return tasks

    @staticmethod
    def expand_dict(**kwargs: Any) -> Iterator[Dict]:
        """Expand a dictionary of iterables into combinations of values.

        Given kwargs and values corresponding to node_args,
        return a dict of combinations of those node_args. Values of kwargs must be iterables.
        """
        # Return empty if no arguments provided
        if len(kwargs) == 0:
            return
        keys = kwargs.keys()
        for element in product(*kwargs.values()):
            yield dict(zip(keys, element))

    def get_tasks_by_node_args(self, **kwargs: Any) -> List["Task"]:
        """Query tasks by node args. Used for setting dependencies."""
        tasks: List["Task"] = []

        for task in self.tasks.values():
            key_count_to_meet = len(kwargs)
            for key, val in kwargs.items():
                if (
                    isinstance(val, Iterable)
                    and task.node.node_args[key] in val
                    or task.node.node_args[key] == val
                ):
                    key_count_to_meet -= 1
                    continue
                else:
                    break
            if key_count_to_meet == 0:
                tasks.append(task)

        return tasks

    def validate(self) -> None:
        # check
        pass

    def bind(self) -> None:
        """Add an array to the database."""
        app_route = "/array"
        rc, resp = self.requester.send_request(
            app_route=app_route,
            message={
                "task_template_version_id": self.task_template_version.id,
                "workflow_id": self.workflow.workflow_id,
                "max_concurrently_running": self.max_concurrently_running,
                "name": self.name,
            },
            request_type="post",
        )

        if rc != StatusCodes.OK:
            raise InvalidResponse(
                f"Unexpected status code {rc} from POST request through route "
                f"{app_route}. Expected code 200. Response content: {resp}"
            )

        array_id = resp["array_id"]
        self.array_id = array_id



"""Task object defines a single executable object that will be added to a Workflow.

TaskInstances will be created from it for every execution.
"""

from __future__ import annotations

import hashlib
from http import HTTPStatus as StatusCodes
import logging
import numbers
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    TYPE_CHECKING,
    Union,
)

from jobmon.client.node import Node
from jobmon.client.task_resources import TaskResources
from jobmon.core.constants import SpecialChars
from jobmon.core.exceptions import InvalidResponse
from jobmon.core.requester import Requester
from jobmon.core.serializers import (
    SerializeTaskInstanceErrorLog,
    SerializeTaskResourceUsage,
)

if TYPE_CHECKING:
    from jobmon.client.array import Array
    from jobmon.client.workflow import Workflow

logger = logging.getLogger(__name__)


def validate_task_resource_scales(resource_scales: Dict[str, Any]) -> None:
    """Validate resource scales are expected types."""
    for scaler in resource_scales.values():
        if not (
            isinstance(scaler, numbers.Number)
            or callable(scaler)
            or isinstance(scaler, Iterator)
        ):
            raise ValueError(
                "Keys in the resource_scales dictionary must be either numeric "
                f"values, Iterators, or Python Callables; found {scaler}, type "
                f"{type(scaler)} instead."
            )


class Task:
    """Task object defines a single executable object that will be added to a Workflow.

    Task Instances will be created from it for every execution.
    """

    @staticmethod
    def is_valid_job_name(name: str) -> bool:
        """If the name is invalid it will raises an exception.

        Primarily based on the restrictions SGE places on job names. The list of illegal
        characters might not be complete, I could not find an official list.

        Must:
          - Not be null or the empty string
          - being with a digit
          - contain am illegal character

        Args:
            name:

        Returns:
            True (or raises)

        Raises:
            ValueError: if the name is not valid.
        """
        if not name:
            raise ValueError("name cannot be None or empty")
        elif name[0].isdigit():
            raise ValueError(f"name cannot begin with a digit, saw: '{name[0]}'")
        elif any(e in name for e in SpecialChars.ILLEGAL_SPECIAL_CHARACTERS):
            raise ValueError(
                f"name contains illegal special character, illegal characters "
                f"are: '{SpecialChars.ILLEGAL_SPECIAL_CHARACTERS}'"
            )
        return True

    def __init__(
        self,
        node: Node,
        task_args: Dict[str, Any],
        op_args: Dict[str, Any],
        array: Optional[Array] = None,
        cluster_name: str = "",
        compute_resources: Optional[Dict[str, Any]] = None,
        compute_resources_callable: Optional[Callable] = None,
        resource_scales: Optional[Dict[str, float]] = None,
        fallback_queues: Optional[List[str]] = None,
        name: Optional[str] = None,
        max_attempts: Optional[int] = None,
        upstream_tasks: Optional[List[Task]] = None,
        task_attributes: Union[List, dict, None] = None,
        requester: Optional[Requester] = None,
    ) -> None:
        """Create a single executable object in the workflow, aka a Task.

        Relate it to a Task Template in order to classify it as a type of job within the
        context of your workflow.

        Args:
            node: Node this task is associated with.
            task_args: Task arguments that make the command unique across workflows
                usually pertaining to data flowing through the task.
            op_args: Task arguments that can change across runs of the same workflow.
                usually pertaining to trivial things like log level or code location.
            array: the array that the task is associated with.
            cluster_name: the name of the cluster the user wants to run their task on.
            compute_resources: A dictionary that includes the users requested resources
                for the current run. E.g. {cores: 1, mem: 1, runtime: 60, queue: all.q}.
            compute_resources_callable: callable compute resources.
            resource_scales: how much users want to scale their resource request if the
                the initial request fails.
            fallback_queues: a list of queues that a user wants to try if their original
                queue is unable to accommodate their requested resources.
            name: name that will be visible in the job status information (e.g. squeue or
                qstat) for this job.
            max_attempts: number of attempts to allow the cluster to try before giving
                up. Default is 3.
            upstream_tasks: Task objects that must be run prior to this
            task_attributes: dictionary of attributes and their values or list
                of attributes that will be assigned later.
            requester: requester object to communicate with the flask services.

        Raise:
            ValueError: If the hashed command is not allowed as an SGE job name; see
                is_valid_job_name
        """
        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

        # pre bind hash defining attributes
        self.node = node
        self.task_args = task_args
        self.mapped_task_args = (
            self.node.task_template_version.convert_arg_names_to_ids(**self.task_args)
        )
        self.task_args_hash = self._hash_task_args()
        self.op_args = op_args

        # pre bind mutable attributes
        self.command = self.node.task_template_version.command_template.format(
            **self.node.node_args, **self.task_args, **self.op_args
        )

        # Not all tasks bound to an array initially
        if array is not None:
            self.array = array

        # Names of jobs can't start with a numeric.
        if not name:
            name = self.node.default_name
        self.is_valid_job_name(name)
        self.name = name

        # upstream and downstream task relationships
        self.upstream_tasks: Set[Task] = (
            set(upstream_tasks) if upstream_tasks else set()
        )
        self.downstream_tasks: Set[Task] = set()
        for task in self.upstream_tasks:
            self.add_upstream(task)

        self.task_attributes: dict = {}
        if isinstance(task_attributes, List):
            for attr in task_attributes:
                self.task_attributes[attr] = None
        elif isinstance(task_attributes, dict):
            for attr in task_attributes:
                self.task_attributes[str(attr)] = str(task_attributes[attr])
        else:
            raise ValueError(
                "task_attributes must be provided as a list of attributes or a "
                "dictionary of attributes and their values"
            )

        # mutable operational/cluster behaviour
        self._instance_max_attempts = max_attempts
        self._instance_cluster_name = cluster_name
        self._instance_compute_resources = (
            compute_resources if compute_resources is not None else {}
        )
        self._instance_compute_resources_callable = compute_resources_callable
        self._instance_resource_scales = (
            resource_scales if resource_scales is not None else {}
        )

        self.fallback_queues: List[str] = (
            fallback_queues if fallback_queues is not None else []
        )

        # error api
        self._errors: Union[
            None, Dict[str, Union[int, List[Dict[str, Union[str, int]]]]]
        ] = None

    @property
    def compute_resources(self) -> Dict[str, Any]:
        try:
            resources = self.array.compute_resources
        except AttributeError:
            resources = {}
        resources.update(self._instance_compute_resources.copy())
        return resources

    @property
    def requested_resources(self) -> Dict[str, Any]:
        """A dictionary that includes the users requested resources for the current run.

        E.g. {cores: 1, mem: 1, runtime: 60, queue: all.q}.
        """
        resources = self.compute_resources
        try:
            resources.pop("queue")
        except KeyError:
            pass
        return resources

    @property
    def resource_scales(self) -> Dict[str, float]:
        """A dictionary that includes the users requested resource scales for the current run.

        E.g. {memory: 0.1, runtime: 0.7}.
        """
        try:
            scales = self.array.resource_scales
        except AttributeError:
            scales = {}
        scales.update(self._instance_resource_scales.copy())
        return scales if scales else {"memory": 0.5, "runtime": 0.5}

    @property
    def cluster_name(self) -> str:
        """The name of the cluster the user wants to run their task on."""
        cluster_name = self._instance_cluster_name
        if not cluster_name:
            try:
                cluster_name = self.array.cluster_name
            except AttributeError:
                # array hasn't been inferred yet. safe to return empty string for now
                pass
        return cluster_name

    @property
    def max_attempts(self) -> int:
        """Get the max_attempts."""
        ma = self._instance_max_attempts
        if not ma:
            try:
                ma = self.array.max_attempts
            except AttributeError:
                # max_attempts hasn't been inferred yet. safe to return empty string for now
                pass
            finally:
                if ma is None:
                    ma = 3
                    self._instance_max_attempts = ma
        return ma

    @property
    def compute_resources_callable(self) -> Optional[Callable]:
        """A callable that returns a compute resources dict."""
        compute_resources_callable = self._instance_compute_resources_callable
        if compute_resources_callable is None:
            compute_resources_callable = self.array.compute_resources_callable
        return compute_resources_callable

    @property
    def queue_name(self) -> str:
        resources = self.compute_resources
        try:
            queue_name = resources.pop("queue")
        except KeyError:
            raise ValueError(
                "A queue name must be provided in the specified compute resources."
            )
        return queue_name

    @property
    def original_task_resources(self) -> TaskResources:
        """Get the id of the task if it has been bound to the db otherwise raise an error."""
        if not hasattr(self, "_original_task_resources"):
            raise AttributeError(
                "task_resources cannot be accessed before workflow is bound"
            )
        return self._original_task_resources

    @original_task_resources.setter
    def original_task_resources(self, val: TaskResources) -> None:
        if not isinstance(val, TaskResources):
            raise ValueError("task_resources must be of type=TaskResources")
        self._original_task_resources = val

    @property
    def is_bound(self) -> bool:
        """If the task template version has been bound to the database."""
        return hasattr(self, "_task_id")

    @property
    def task_id(self) -> int:
        """Get the id of the task if it has been bound to the db otherwise raise an error."""
        if not self.is_bound:
            raise AttributeError("task_id cannot be accessed before task is bound")
        return self._task_id

    @task_id.setter
    def task_id(self, val: int) -> None:
        self._task_id = val

    @property
    def initial_status(self) -> str:
        """Get initial status of the task if it has been bound to the db; else raise error."""
        if not hasattr(self, "_initial_status"):
            raise AttributeError(
                "initial_status cannot be accessed before task is bound"
            )
        return self._initial_status

    @initial_status.setter
    def initial_status(self, val: str) -> None:
        self._initial_status = val

    @property
    def final_status(self) -> str:
        """Get initial status of the task if it has been bound, otherwise raise error."""
        if not hasattr(self, "_final_status"):
            raise AttributeError(
                "final_status cannot be accessed until workflow is run"
            )
        return self._final_status

    @final_status.setter
    def final_status(self, val: str) -> None:
        self._final_status = val

    @property
    def array(self) -> Array:
        """Get the array the task has been added to or else raise an AttributeError."""
        if not hasattr(self, "_array"):
            raise AttributeError(
                "array cannot be accessed via task before task is added to array"
            )
        return self._array

    @array.setter
    def array(self, val: Array) -> None:
        self._array = val

    @property
    def workflow(self) -> Workflow:
        """Get the workflow the task has been added to or else raise an AttributeError."""
        if not hasattr(self, "_workflow"):
            raise AttributeError(
                "workflow cannot be accessed via task before workflow is added to workflow"
            )
        return self._workflow

    @workflow.setter
    def workflow(self, val: Workflow) -> None:
        self._workflow = val

    def add_upstream(self, ancestor: Task) -> None:
        """Add an upstream (ancestor) Task.

        This has Set semantics, an upstream task will only be added once. Symmetrically, this
        method also adds this Task as a downstream on the ancestor.
        """
        self.upstream_tasks.add(ancestor)
        ancestor.downstream_tasks.add(self)

        self.node.add_upstream_node(ancestor.node)

    def add_upstreams(self, tasks: List[Task]) -> None:
        """Add all Tasks in user provided list as upstreams."""
        for task in tasks:
            self.add_upstream(task)

    def add_downstream(self, descendent: Task) -> None:
        """Add a downstream (ancestor) Task.

        This has Set semantics, a downstream task will only be added once. Symmetrically,
        this method also adds this Task as an upstream on the ancestor.
        """
        self.downstream_tasks.add(descendent)
        descendent.upstream_tasks.add(self)

        self.node.add_downstream_node(descendent.node)

    def add_downstreams(self, tasks: List[Task]) -> None:
        """Add all Tasks in user provided list as downstreams."""
        for task in tasks:
            self.add_downstream(task)

    def add_attribute(self, attribute: str, value: str) -> None:
        """Function that users can call to add a single attribute for a task."""
        self.task_attributes[str(attribute)] = str(value)

    def get_errors(
        self,
    ) -> Union[None, Dict[str, Union[int, List[Dict[str, Union[str, int]]]]]]:
        """Return all errors for each task, with the recent task_instance_id actually used."""
        if (
            self._errors is None
            and hasattr(self, "_task_id")
            and self._task_id is not None
        ):
            return_code, response = self.requester.send_request(
                app_route=f"/task/{self._task_id}/most_recent_ti_error",
                message={},
                request_type="get",
            )
            if return_code == StatusCodes.OK:
                task_instance_id = response["task_instance_id"]
                if task_instance_id is not None:
                    rc, response = self.requester.send_request(
                        app_route=f"/task_instance/{task_instance_id}"
                        f"/task_instance_error_log",
                        message={},
                        request_type="get",
                    )
                    errors_ti = [
                        SerializeTaskInstanceErrorLog.kwargs_from_wire(j)
                        for j in response["task_instance_error_log"]
                    ]
                    self._errors = {
                        "task_instance_id": task_instance_id,
                        "error_log": errors_ti,
                    }

        return self._errors

    def set_compute_resources_from_yaml(
        self, cluster_name: str, yaml_file: str
    ) -> None:
        """Set default compute resources from a user provided yaml file for task level.

        TODO: Implement this method.

        Args:
            cluster_name: name of cluster to set default values for.
            yaml_file: the yaml file that is providing the compute resource values.
        """
        pass

    def update_compute_resources(self, **kwargs: Any) -> None:
        """Function that allows users to update their compute resources."""
        self.compute_resources.update(kwargs)

    def update_resource_scales(self, **kwargs: Any) -> None:
        """Function that allows users to update their resource scales."""
        self.resource_scales.update(kwargs)

    def _hash_task_args(self) -> int:
        """A hash of the encoded result of the args and values concatenated together."""
        arg_ids = list(self.mapped_task_args.keys())
        arg_ids.sort()

        arg_values = [str(self.mapped_task_args[key]) for key in arg_ids]
        str_arg_ids = [str(arg) for arg in arg_ids]

        hash_value = int(
            hashlib.sha256(
                "".join(str_arg_ids + arg_values).encode("utf-8")
            ).hexdigest(),
            16,
        )
        return hash_value

    def __eq__(self, other: object) -> bool:
        """Check if the hashes of two tasks are equivalent."""
        if not isinstance(other, Task):
            return False
        else:
            return hash(self) == hash(other)

    def __lt__(self, other: Task) -> bool:
        """Check if one hash is less than the has of another Task."""
        return hash(self) < hash(other)

    def __hash__(self) -> int:
        """Create the hash for a task to determine if it is unique within a dag."""
        if not hasattr(self, "_hash_val"):
            hash_value = hashlib.sha256()
            hash_value.update(bytes(str(hash(self.node)).encode("utf-8")))
            hash_value.update(bytes(str(self.task_args_hash).encode("utf-8")))
            self._hash_val = int(hash_value.hexdigest(), 16)
        return self._hash_val

    def __repr__(self) -> str:
        """A representation string for a Task instance."""
        repr_string = (
            f"Task(command={self.command}, "
            f"name={self.name}, "
            f"node={self.node}, "
            f"task_args={self.task_args}"
        )
        try:
            repr_string += f", task_id={self.task_id})"
        except AttributeError:
            repr_string += ")"
        return repr_string

    def resource_usage(self) -> dict:
        """Get the resource usage for the successful TaskInstance of a Task."""
        app_route = "/task_resource_usage"
        return_code, response = self.requester.send_request(
            app_route=app_route,
            message={"task_id": self.task_id},
            request_type="get",
        )
        if return_code != StatusCodes.OK:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from GET "
                f"request through route {app_route}. Expected code "
                f"200. Response content: {response}"
            )
        return SerializeTaskResourceUsage.kwargs_from_wire(response)



"""Task Templates are versioned to recognize changes to args and command templates."""

from __future__ import annotations

import hashlib
from http import HTTPStatus as StatusCodes
import logging
from string import Formatter
from typing import Any, Dict, List, Optional, Tuple, Type, TYPE_CHECKING

from jobmon.core.exceptions import InvalidResponse
from jobmon.core.requester import Requester
from jobmon.core.serializers import SerializeClientTaskTemplateVersion

if TYPE_CHECKING:
    from jobmon.client.task_template import TaskTemplate

logger = logging.getLogger(__name__)


class TaskTemplateVersion:
    """Task Templates are versioned to recognize changes to args and command templates."""

    def __init__(
        self,
        command_template: str,
        node_args: list,
        task_args: list,
        op_args: list,
        requester: Optional[Requester] = None,
    ) -> None:
        """Initialization of task template version object."""
        # id vars
        self.command_template = command_template

        # hash args
        self._node_args: set
        self.node_args = set(node_args)
        self._task_args: set
        self.task_args = set(task_args)
        self._op_args: set
        self.op_args = set(op_args)

        # binding attributes
        self._task_template_version_id: int
        self._id_name_map: Dict
        self._task_template: TaskTemplate

        self.default_compute_resources_set: Dict[str, Dict[str, Any]] = {}
        self.default_resource_scales_set: Dict[str, Dict[str, float]] = {}
        self.default_cluster_name: str = ""
        self.default_max_attempts: Optional[int] = None

        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

    @classmethod
    def get_task_template_version(
        cls: Type[TaskTemplateVersion],
        task_template: TaskTemplate,
        command_template: str,
        node_args: List[str] = [],
        task_args: List[str] = [],
        op_args: List[str] = [],
    ) -> TaskTemplateVersion:
        """Get a bound TaskTemplateVersion object from parameters.

        Args:
            task_template: TaskTemplate this version should be associated with.
            command_template: an abstract command representing a task, where the arguments to
                the command have defined names but the values are not assigned. eg: '{python}
                {script} --data {data} --para {para} {verbose}'
            node_args: any named arguments in command_template that make the command unique
                within this template for a given workflow run. Generally these are arguments
                that can be parallelized over.
            task_args: any named arguments in command_template that make the command unique
                across workflows if the node args are the same as a previous workflow.
                Generally these are arguments about data moving though the task.
            op_args: any named arguments in command_template that can change without changing
                the identity of the task. Generally these are things like the task executable
                location or the verbosity of the script.
        """
        task_template_version = cls(
            command_template,
            node_args,
            task_args,
            op_args,
            task_template.requester,
        )
        task_template_version.bind(task_template)
        return task_template_version

    @classmethod
    def from_wire(
        cls: Type[TaskTemplateVersion], wire_tuple: Tuple, task_template: TaskTemplate
    ) -> TaskTemplateVersion:
        """Get a bound TaskTemplateVersion object from the http wire format.

        Args:
            task_template: TaskTemplate this version should be associated with.
            wire_tuple: Wire format for ToolVersion defined in jobmon.serializers.
        """
        kwargs = SerializeClientTaskTemplateVersion.kwargs_from_wire(wire_tuple)

        # post bind args should be popped off and added as attrs
        task_template_version_id = kwargs.pop("task_template_version_id")
        id_name_map = kwargs.pop("id_name_map")
        task_template_id = kwargs.pop("task_template_id")
        if task_template_id != task_template.id:
            raise ValueError(
                "task_template_id from wire_tuple does not match task_template. "
                f"Expected {task_template.id} from wire_tuple. Got "
                f"{kwargs['task_template_id']}"
            )

        # instantiate and add attrs
        task_template_version = cls(
            command_template=kwargs["command_template"],
            node_args=kwargs["node_args"],
            task_args=kwargs["task_args"],
            op_args=kwargs["op_args"],
            requester=task_template.requester,
        )
        task_template_version._task_template = task_template
        task_template_version._task_template_version_id = task_template_version_id
        task_template_version._id_name_map = id_name_map
        return task_template_version

    def bind(self, task_template: TaskTemplate) -> None:
        """Bind task template version to the DB.

        Args:
            task_template: the TaskTemplate that this version is associated with.
        """
        if self.is_bound:
            return

        app_route = f"/task_template/{task_template.id}/add_version"
        return_code, response = self.requester.send_request(
            app_route=app_route,
            message={
                "command_template": self.command_template,
                "arg_mapping_hash": self.arg_mapping_hash,
                "node_args": list(self.node_args),
                "task_args": list(self.task_args),
                "op_args": list(self.op_args),
            },
            request_type="post",
        )

        if return_code != StatusCodes.OK:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST request through route "
                f"{app_route}. Expected code 200. Response content: {response}"
            )

        response_dict = SerializeClientTaskTemplateVersion.kwargs_from_wire(
            response["task_template_version"]
        )

        self._task_template = task_template
        self._task_template_version_id = response_dict["task_template_version_id"]
        self._id_name_map = response_dict["id_name_map"]

    @property
    def is_bound(self) -> bool:
        """If the task template version has been bound to the database."""
        return hasattr(self, "_task_template_version_id")

    @property
    def task_template(self) -> TaskTemplate:
        if not self.is_bound:
            raise AttributeError(
                "task_template cannot be accessed before TaskTemplateVersion is bound"
            )
        return self._task_template

    @property
    def id(self) -> int:
        """The unique ID of the task template version if it has been bound."""
        if not self.is_bound:
            raise AttributeError(
                "id cannot be accessed before TaskTemplateVersion is bound"
            )
        return self._task_template_version_id

    @property
    def id_name_map(self) -> Dict[str, int]:
        """Map of arg ids to arg names if bound to the db."""
        if not self.is_bound:
            raise AttributeError(
                "arg_id_name_map cannot be accessed before TaskTemplateVersion is bound"
            )
        return self._id_name_map

    @property
    def template_args(self) -> set:
        """The argument names in the command template."""
        return set(
            [i[1] for i in Formatter().parse(self.command_template) if i[1] is not None]
        )

    @property
    def node_args(self) -> set:
        """Return task template version node args.

        Any named arguments in command_template that make the command unique within this
        template for a given workflow run. Generally these are arguments that can be
        parallelized over.
        """
        return self._node_args

    @node_args.setter
    def node_args(self, val: set) -> None:
        """Set the node args."""
        if self.is_bound:
            raise AttributeError(
                "Cannot set node_args. node_args must be declared during "
                "instantiation"
            )
        if "name" in val:
            raise ValueError("Name is not allowed as a keyword in a command_template.")
        if not self.template_args.issuperset(val):
            raise ValueError(
                "The format keys declared in command_template must be a "
                "superset of the keys declared in node_args. Values received "
                f"were --- \ncommand_template is: {self.command_template}. "
                f"\ncommand_template format keys are {self.template_args}. "
                f"\nnode_args is: {val}. \nmissing format keys in "
                f"command_template are {set(val) - self.template_args}."
            )
        self._node_args = val

    @property
    def task_args(self) -> set:
        """Task template version task args.

        Any named arguments in command_template that make the command unique
        across workflows if the node args are the same as a previous workflow.
        Generally these are arguments about data moving though the task.
        """
        return self._task_args

    @task_args.setter
    def task_args(self, val: set) -> None:
        """Set the task args."""
        if self.is_bound:
            raise AttributeError(
                "Cannot set task_args. task_args must be declared during "
                "instantiation"
            )
        if "name" in val:
            raise ValueError("Name is not allowed as a keyword in a command_template.")
        if not self.template_args.issuperset(val):
            raise ValueError(
                "The format keys declared in command_template must bes a "
                "superset of the keys declared in task_args. Values received "
                f"were --- \ncommand_template is: {self.command_template}. "
                f"\ncommand_template format keys are {self.template_args}. "
                f"\ntask_args is: {val}. \nmissing format keys in "
                f"command_template are {set(val) - self.template_args}."
            )
        self._task_args = val

    @property
    def op_args(self) -> set:
        """Return the task template version OP args.

        Any named arguments in command_template that can change without changing the
        identity of the task. Generally these are things like the task executable location or
        the verbosity of the script.
        """
        return self._op_args

    @op_args.setter
    def op_args(self, val: set) -> None:
        """Setting op args."""
        if self.is_bound:
            raise AttributeError(
                "Cannot set op_args. op_args must be declared during " "instantiation"
            )
        if "name" in val:
            raise ValueError("Name is not allowed as a keyword in a command_template.")
        if not self.template_args.issuperset(val):
            raise ValueError(
                "The format keys declared in command_template must be a "
                "superset of the keys declared in op_args. Values received "
                f"were --- \ncommand_template is: {self.command_template}. "
                f"\ncommand_template format keys are {self.template_args}. "
                f"\nop_args is: {val}. \nmissing format keys in "
                f"command_template are {set(val) - self.template_args}."
            )
        self._op_args = val

    @property
    def arg_mapping_hash(self) -> int:
        """Hash args to identify unique task_template."""
        node_args = "".join(sorted(self.node_args))
        task_args = "".join(sorted(self.task_args))
        op_args = "".join(sorted(self.op_args))
        hashable = ",".join([node_args, task_args, op_args])
        return int(hashlib.sha256(hashable.encode("utf-8")).hexdigest(), 16)

    def filter_kwargs(self, arg_type: str, **kwargs: str) -> Dict[str, Any]:
        """Return the set of kwargs that are of arg_type.

        Args:
            arg_type: either node_args, task_args, op_args
            kwargs: the key/value pairs to be filtered by type
        """
        arg_type_set_map = {
            "node_args": self.node_args,
            "task_args": self.task_args,
            "op_args": self.op_args,
        }
        arg_set = arg_type_set_map[arg_type]
        result = {}
        for key, val in kwargs.items():
            if key in arg_set:
                result[key] = val
        return result

    def convert_arg_names_to_ids(self, **kwargs: str) -> Dict[int, Any]:
        """Map from names to ids."""
        return {self.id_name_map[k]: str(v) for k, v in kwargs.items()}

    def update_default_compute_resources(
        self, cluster_name: str, **kwargs: Any
    ) -> None:
        """Update compute resources in place only overridding specified keys.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to modify default values for.
            **kwargs: any key/value pair you want to update specified as an argument.
        """
        compute_resources = {cluster_name: kwargs}
        self.default_compute_resources_set.update(compute_resources)

    def update_default_resource_scales(self, cluster_name: str, **kwargs: Any) -> None:
        """Update resource scales in place only overridding specified keys.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to modify default values for.
            **kwargs: any key/value pair you want to update specified as an argument.
        """
        resource_scales = {cluster_name: kwargs}
        self.default_resource_scales_set.update(resource_scales)

    def set_default_compute_resources_from_dict(
        self, cluster_name: str, compute_resources: Dict[str, Any]
    ) -> None:
        """Set compute resources for a given cluster_name.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to set default values for.
            compute_resources: dictionary of default compute resources to run tasks
                with. Can be overridden at task template or task level.
                dict of {resource_name: resource_value}
        """
        self.default_compute_resources_set[cluster_name] = compute_resources

    def set_default_resource_scales_from_dict(
        self, cluster_name: str, resource_scales: Dict[str, float]
    ) -> None:
        """Set compute resources and scales for a given cluster_name.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to set default values for.
            resource_scales: dictionary of default resource scales to adjust task
                resources with. Can be overridden at task level.
                dict of {resource_name: scale_value}
        """
        self.default_resource_scales_set[cluster_name] = resource_scales

    def set_default_max_attempts(self, value: int) -> None:
        """Set default max attempts at tool leve.

        Args:
            value: the default max attempts value.
        """
        self.default_max_attempts = value

    def __hash__(self) -> int:
        """Unique identifier for this object."""
        hash_value = hashlib.sha256()
        hash_value.update(bytes(str(self.arg_mapping_hash).encode("utf-8")))
        hash_value.update(bytes(str(self.command_template).encode("utf-8")))
        return int(hash_value.hexdigest(), 16)

    def __repr__(self) -> str:
        """A representation string for a TaskTemplateVersion instance."""
        repr_string = (
            f"TaskTemplateVersion(command_template={self.command_template}, "
            f"node_args={self.node_args}, "
            f"task_args={self.task_args}, "
            f"op_args={self.op_args}"
        )
        try:
            repr_string += f", id={self.id}"
        except AttributeError:
            repr_string += ")"

        return repr_string



"""Array object used by swarm to create task instance batches."""

from __future__ import annotations

import logging
from typing import Set

from jobmon.client.swarm.swarm_task import SwarmTask


logger = logging.getLogger(__name__)


class SwarmArray:
    def __init__(self, array_id: int, max_concurrently_running: int) -> None:
        """Initialization of the SwarmArray."""
        self.array_id = array_id
        self.tasks: Set[SwarmTask] = set()
        self.max_concurrently_running = max_concurrently_running

    def add_task(self, task: SwarmTask) -> None:
        if task.array_id != self.array_id:
            raise ValueError(
                f"array_id mismatch. SwarmTask={task.array_id}. Array={self.array_id}."
            )
        self.tasks.add(task)

    def __hash__(self) -> int:
        """Returns the array ID."""
        return self.array_id

    def __eq__(self, other: object) -> bool:
        """Check if the hashes of two arrays are equivalent."""
        if not isinstance(other, SwarmArray):
            return False
        else:
            return hash(self) == hash(other)

    def __lt__(self, other: SwarmArray) -> bool:
        """Check if one hash is less than the has of another DistributorArray."""
        return hash(self) < hash(other)



"""Client command line interface for workflow/task status and concurrency limiting."""

import argparse
import json
from typing import Any, Optional

from jobmon.client.status_commands import get_task_dependencies
from jobmon.core.cli import CLI


class _HelpAction(argparse._HelpAction):
    """To show help for all subparsers in one place."""

    def __call__(
        self,
        parser: Any,
        namespace: argparse.Namespace,
        values: Any,
        option_string: Optional[str] = None,
    ) -> None:
        """Add subparsers' help info when jobmon --help is called."""
        print(parser.format_help())
        subparsers_actions = [
            action
            for action in parser._actions
            if isinstance(action, argparse._SubParsersAction)
        ]
        print("Jobmon Usage Options:")
        for sub_action in subparsers_actions:
            for choice, subparser in sub_action.choices.items():
                print(
                    f"{choice.upper()} (for more information specify 'jobmon {choice} "
                    f"--help'):"
                )
                print(subparser.format_usage())
        parser.exit()


class ClientCLI(CLI):
    """Client command line interface for workflow/task status and concurrency limiting."""

    def __init__(self) -> None:
        """Initialization of client CLI."""
        self.parser = argparse.ArgumentParser("Jobmon Client CLI", add_help=False)
        self.parser.add_argument(
            "--help", action=_HelpAction, help="Help if you need Help"
        )
        self._subparsers = self.parser.add_subparsers(
            dest="sub_command", parser_class=argparse.ArgumentParser
        )

        self._add_workflow_status_subparser()
        self._add_workflow_tasks_subparser()
        self._add_task_template_resources_subparser()
        self._add_task_status_subparser()
        self._add_update_task_status_subparser()
        self._add_concurrency_limit_subparser()
        self._add_version_subparser()
        self._add_task_dependencies_subparser()
        self._add_workflow_reset_subparser()
        self._add_create_resource_yaml_subparser()
        self._add_get_filepaths_subparser()
        self._add_resume_workflow_parser()

    @staticmethod
    def limit_checker(limit: Any) -> int:
        """Coerce to int and check that the limit is greater than 0."""
        limit = int(limit)
        if limit <= 0:
            raise argparse.ArgumentTypeError(
                f"Limit value must be greater than 0. The limit "
                f"value passed in was: {limit}"
            )
        return limit

    @staticmethod
    def workflow_status(args: argparse.Namespace) -> None:
        """Workflow status checking options."""
        from tabulate import tabulate
        from jobmon.client.status_commands import workflow_status as workflow_status_cmd

        df = workflow_status_cmd(args.workflow_id, args.user, args.json, args.limit)
        if args.json:
            print(df)
        else:
            print(tabulate(df, headers="keys", tablefmt="psql", showindex=False))

    @staticmethod
    def workflow_tasks(args: argparse.Namespace) -> None:
        """Check the tasks for a given workflow."""
        from tabulate import tabulate
        from jobmon.client.status_commands import workflow_tasks as workflow_tasks_cmd

        df = workflow_tasks_cmd(args.workflow_id, args.status, args.json, args.limit)
        if args.json:
            print(df)
        else:
            print(tabulate(df, headers="keys", tablefmt="psql", showindex=False))

    @staticmethod
    def task_template_resources(args: argparse.Namespace) -> None:
        """Aggregates the resource usage for a given TaskTemplateVersion."""
        from jobmon.client.status_commands import task_template_resources

        resources = task_template_resources(
            args.task_template_version, args.workflows, args.node_args
        )
        print(resources)

    @staticmethod
    def task_status(args: argparse.Namespace) -> None:
        """Check task status."""
        from tabulate import tabulate
        from jobmon.client.status_commands import task_status as task_status_cmd

        df = task_status_cmd(args.task_ids, args.status, args.json)
        print(f"\nTASK_IDS: {args.task_ids}")
        if args.json:
            print(df)
        else:
            print(tabulate(df, headers="keys", tablefmt="psql", showindex=False))

    @staticmethod
    def update_task_status(args: argparse.Namespace) -> None:
        """Manually update task status for resumes, reruns, etc."""
        from jobmon.client.status_commands import update_task_status

        response = update_task_status(
            args.task_ids,
            args.workflow_id,
            args.new_status,
            args.force,
            args.recursive,
        )
        print(f"Response is: {response}")

    @staticmethod
    def concurrency_limit(args: argparse.Namespace) -> None:
        """Set a limit for the number of tasks that can run concurrently."""
        from jobmon.client.status_commands import (
            concurrency_limit as concurrency_limit_cmd,
        )

        response = concurrency_limit_cmd(args.workflow_id, args.max_tasks)
        print(response)

    @staticmethod
    def task_dependencies(args: argparse.Namespace) -> None:
        """Get task's upstream and downstream tasks and their status."""
        r = get_task_dependencies(args.task_id)
        up = r["up"]
        down = r["down"]
        """Format output that should look like:
        Upstream Tasks:

             Task ID         Status
             1               D

        Downstream Tasks:

            Task ID         Status
            3               D
            4               D
        """
        print("Upstream Tasks:\n")
        print("{:<8} {:<15} {:<15}".format("", "Task ID", "Status"))
        for item in up:
            task_id = item["id"]
            status = item["status"]
            print("{:<8} {:<15} {:<15}".format("", task_id, status))
        print("\nDownstream Tasks:\n")
        print("{:<8} {:<15} {:<15}".format("", "Task ID", "Status"))
        for item in down:
            task_id = item["id"]
            status = item["status"]
            print("{:<8} {:<15} {:<15}".format("", task_id, status))

    @staticmethod
    def workflow_reset(args: argparse.Namespace) -> None:
        """Manually reset a workflow."""
        from jobmon.client.status_commands import workflow_reset

        response = workflow_reset(
            workflow_id=args.workflow_id, partial_reset=args.partial_reset
        )
        print(f"Response is: {response}")

    @staticmethod
    def jobmon_version(args: argparse.Namespace) -> None:
        """Return the jobmon version."""
        from jobmon.client import __version__

        print(__version__)

    @staticmethod
    def resource_yaml(args: argparse.Namespace) -> None:
        """Create resource yaml."""
        from jobmon.client.status_commands import create_resource_yaml

        # input check
        if (args.workflow_id is None) ^ (args.task_id is None):
            r = create_resource_yaml(
                args.workflow_id,
                args.task_id,
                args.value_mem,
                args.value_core,
                args.value_runtime,
                args.clusters,
            )
            if args.print:
                print(r)
            if args.file:
                f = open(args.file, "w")
                f.write(r)
                f.close()
        else:
            print("Please provide a value for either -w or -t but not both.")

    @staticmethod
    def resume_workflow(args: argparse.Namespace) -> None:
        """Resume a workflow from a workflow ID."""
        from jobmon.client.status_commands import resume_workflow_from_id

        resume_workflow_from_id(
            workflow_id=args.workflow_id,
            cluster_name=args.cluster_name,
            reset_if_running=args.reset_running_jobs,
            timeout=args.timeout,
        )

    @staticmethod
    def get_filepaths(args: argparse.Namespace) -> None:
        from tabulate import tabulate
        from jobmon.client.status_commands import get_filepaths

        df = get_filepaths(
            workflow_id=args.workflow_id,
            array_name=args.array_name,
            job_name=args.job_name,
            limit=args.limit,
        )
        if args.json:
            print(df)
        else:
            print(tabulate(df, headers="keys", tablefmt="psql", showindex=False))

    def _add_version_subparser(self) -> None:
        version_parser = self._subparsers.add_parser("version")
        version_parser.set_defaults(func=self.jobmon_version)

    def _add_workflow_status_subparser(self) -> None:
        workflow_status_parser = self._subparsers.add_parser("workflow_status")
        workflow_status_parser.set_defaults(func=self.workflow_status)
        workflow_status_parser.add_argument(
            "-w",
            "--workflow_id",
            nargs="*",
            help="list of workflow_ids",
            required=False,
            type=int,
            action="append",
            default=[],
        )
        workflow_status_parser.add_argument(
            "-u", "--user", nargs="*", help="list of users", required=False, type=str
        )
        workflow_status_parser.add_argument(
            "-n", "--json", dest="json", action="store_true"
        )
        workflow_status_parser.add_argument(
            "-l",
            "--limit",
            default=5,
            help="limit the number of returning records; default is 5",
            required=False,
            type=self.limit_checker,
        )

    def _add_workflow_tasks_subparser(self) -> None:
        workflow_tasks_parser = self._subparsers.add_parser("workflow_tasks")
        workflow_tasks_parser.set_defaults(func=self.workflow_tasks)
        workflow_tasks_parser.add_argument(
            "-w",
            "--workflow_id",
            help="workflow_id to get task statuses for",
            required=True,
            type=int,
        )
        workflow_tasks_parser.add_argument(
            "-s",
            "--status",
            nargs="*",
            help="limit tasks to a status (PENDING, RUNNING, DONE, FATAL)",
            choices=[
                "PENDING",
                "RUNNING",
                "DONE",
                "FATAL",
                "pending",
                "running",
                "done",
                "fatal",
            ],
            required=False,
        )
        workflow_tasks_parser.add_argument(
            "-n", "--json", dest="json", action="store_true"
        )
        workflow_tasks_parser.add_argument(
            "-l",
            "--limit",
            default=5,
            help="limit the number of returning records; default is 5",
            required=False,
            type=self.limit_checker,
        )

    def _add_task_template_resources_subparser(self) -> None:
        tt_resources_parser = self._subparsers.add_parser("task_template_resources")
        tt_resources_parser.set_defaults(func=self.task_template_resources)
        tt_resources_parser.add_argument(
            "-t",
            "--task_template_version",
            help="TaskTemplateVersion ID to get resource usage for",
            required=True,
            type=int,
        )
        tt_resources_parser.add_argument(
            "-w",
            "--workflows",
            nargs="*",
            help="list of workflow IDs to query by",
            required=False,
            type=int,
        )
        tt_resources_parser.add_argument(
            "-a",
            "--node_args",
            help="dictionary of node arguments to query by",
            required=False,
            type=json.loads,
        )

    def _add_task_status_subparser(self) -> None:
        task_status_parser = self._subparsers.add_parser("task_status")
        task_status_parser.set_defaults(func=self.task_status)
        task_status_parser.add_argument(
            "-t",
            "--task_ids",
            nargs="+",
            help="task_ids to get task statuses for",
            required=True,
            type=int,
        )
        task_status_parser.add_argument(
            "-s",
            "--status",
            nargs="*",
            help="limit task instances to statuses (PENDING, RUNNING, DONE, FATAL)",
            choices=[
                "PENDING",
                "RUNNING",
                "DONE",
                "FATAL",
                "pending",
                "running",
                "done",
                "fatal",
            ],
            required=False,
        )
        task_status_parser.add_argument(
            "-n", "--json", dest="json", action="store_true"
        )

    def _add_update_task_status_subparser(self) -> None:
        update_task_parser = self._subparsers.add_parser("update_task_status")
        update_task_parser.set_defaults(func=self.update_task_status)
        update_task_parser.add_argument(
            "-t",
            "--task_ids",
            nargs="+",
            help="task_ids to reset",
            required=True,
            type=int,
        )
        update_task_parser.add_argument(
            "-w",
            "--workflow_id",
            help="workflow_id of the tasks to reset",
            required=True,
            type=int,
        )
        update_task_parser.add_argument(
            "-s",
            "--new_status",
            help='Status to set to. "D" for DONE; "G" for REGISTERED/(pending).',
            choices=["D", "G"],
            type=str,
        )
        update_task_parser.add_argument(
            "-f",
            "--force",
            help="If set, allow all source statuses and all workflow statuses.",
            default=False,
            action="store_true",
        )
        update_task_parser.add_argument(
            "-r",
            "--recursive",
            help="If used with --force, Jobmon will apply recursive update_status downstream "
            "or upstream depending on new_status ",
            default=False,
            action="store_true",
        )

    def _add_concurrency_limit_subparser(self) -> None:
        concurrency_limit_parser = self._subparsers.add_parser("concurrency_limit")
        concurrency_limit_parser.set_defaults(func=self.concurrency_limit)
        concurrency_limit_parser.add_argument(
            "-w",
            "--workflow_id",
            required=True,
            type=int,
            help="Workflow ID of the workflow to be adjusted",
        )

        # Define a custom function to validate the user's input.
        def _validate_ntasks(x: Any) -> int:
            try:
                x = int(x)
            except ValueError:
                raise argparse.ArgumentTypeError(f"{x} is not coercible to an integer.")
            if x < 0:
                raise argparse.ArgumentTypeError(
                    "Max concurrent tasks must be at least 0."
                )
            return x

        concurrency_limit_parser.add_argument(
            "-m",
            "--max_tasks",
            required=True,
            type=_validate_ntasks,
            help="Number of concurrent tasks to allow. Must be at least 1.",
        )

    def _add_task_dependencies_subparser(self) -> None:
        task_dependencies_parser = self._subparsers.add_parser("task_dependencies")
        task_dependencies_parser.set_defaults(func=self.task_dependencies)
        task_dependencies_parser.add_argument(
            "-t", "--task_id", help="list of task dependencies", required=True, type=int
        )

    def _add_workflow_reset_subparser(self) -> None:
        workflow_reset_parser = self._subparsers.add_parser("workflow_reset")
        workflow_reset_parser.set_defaults(func=self.workflow_reset)
        workflow_reset_parser.add_argument(
            "-w", "--workflow_id", help="workflow_id to reset", required=True, type=int
        )
        workflow_reset_parser.add_argument(
            "-p",
            "--partial_reset",
            help="Set to indicate Done tasks will not be reset",
            required=False,
            action="store_true",
        )

    def _add_create_resource_yaml_subparser(self) -> None:
        create_resource_yaml_parser = self._subparsers.add_parser(
            "create_resource_yaml"
        )
        create_resource_yaml_parser.set_defaults(func=self.resource_yaml)
        create_resource_yaml_parser.add_argument(
            "-w",
            "--workflow_id",
            help="The workflow id to generate resource YAML. "
            "Must provide either -w or -t.",
            required=False,
            type=int,
        )
        create_resource_yaml_parser.add_argument(
            "-t",
            "--task_id",
            help="The workflow id to generate resource YAML. "
            "Must provide either -w or -t.",
            required=False,
            type=int,
        )
        create_resource_yaml_parser.add_argument(
            "--value_mem",
            help="The algorithm to get memory usage. Default avg.",
            choices=["avg", "max", "min"],
            required=False,
            default="avg",
            type=str,
        )
        create_resource_yaml_parser.add_argument(
            "--value_core",
            help="The algorithm to get core requested. Default avg.",
            choices=["avg", "max", "min"],
            required=False,
            default="avg",
            type=str,
        )
        create_resource_yaml_parser.add_argument(
            "--value_runtime",
            help="The algorithm to get runtime. Default max.",
            choices=["avg", "max", "min"],
            required=False,
            default="max",
            type=str,
        )
        create_resource_yaml_parser.add_argument(
            "-f",
            "--file",
            help="The file to save the YAML.",
            required=False,
            default=None,
            type=str,
        )
        create_resource_yaml_parser.add_argument(
            "-p",
            "--print",
            help="Print the result YAMl to standard output.",
            required=False,
            default=False,
            action="store_true",
        )
        create_resource_yaml_parser.add_argument(
            "-c",
            "--clusters",
            nargs="+",
            help="The clusters for the YAML.",
            required=False,
            default=["slurm"],
            type=str,
        )

    def _add_get_filepaths_subparser(self) -> None:
        get_filepaths_parser = self._subparsers.add_parser("get_filepaths")
        get_filepaths_parser.set_defaults(func=self.get_filepaths)
        get_filepaths_parser.add_argument(
            "-w",
            "--workflow_id",
            help="workflow_id to filter by",
            required=True,
            type=int,
        )
        get_filepaths_parser.add_argument(
            "-a",
            "--array_name",
            help="array name to filter by",
            required=False,
            default="",
            type=str,
        )
        get_filepaths_parser.add_argument(
            "-j",
            "--job_name",
            help="job name to filter by",
            required=False,
            default="",
            type=str,
        )
        get_filepaths_parser.add_argument(
            "-l",
            "--limit",
            default=5,
            help="Limit the number of returning records; default is 5",
            required=False,
            type=self.limit_checker,
        )
        get_filepaths_parser.add_argument(
            "-n", "--json", dest="json", action="store_true"
        )

    def _add_resume_workflow_parser(self) -> None:
        workflow_resume_parser = self._subparsers.add_parser("workflow_resume")
        workflow_resume_parser.set_defaults(func=self.resume_workflow)
        workflow_resume_parser.add_argument(
            "-w", "--workflow_id", help="workflow_id to resume", required=True, type=int
        )
        # TODO: perhaps provide a mechanism to infer the last cluster this
        # workflow was run on
        workflow_resume_parser.add_argument(
            "-c",
            "--cluster_name",
            help="cluster to run this workflow on, e.g. 'slurm', 'slurm_test', 'dummy'",
            required=True,
        )
        workflow_resume_parser.add_argument(
            "--reset-running-jobs",
            help="whether to reset running jobs or not",
            required=False,
            action="store_true",
        )
        workflow_resume_parser.add_argument(
            "-t",
            "--timeout",
            help="timeout for resume command",
            required=False,
            default=180,
        )


def main(argstr: Optional[str] = None) -> None:
    """Create CLI."""
    cli = ClientCLI()
    cli.main(argstr)



"""Configuration setting for client-side only."""

from __future__ import annotations

import logging
import logging.config
import sys
from typing import Dict, Optional, Type


_DEFAULT_LOG_FORMAT = (
    "%(asctime)s [%(name)-12s] %(module)s %(levelname)-8s: %(message)s"
)

default_config: Dict = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {"format": _DEFAULT_LOG_FORMAT, "datefmt": "%Y-%m-%d %H:%M:%S"}
    },
    "handlers": {
        "default": {
            "level": "INFO",
            "class": "logging.StreamHandler",
            "formatter": "default",
            "stream": sys.stdout,
        }
    },
}


class JobmonLoggerConfig:
    """A class to automatically format and attach handlers to client logging modules."""

    dict_config = default_config

    @classmethod
    def set_default_config(
        cls: Type[JobmonLoggerConfig], dict_config: Optional[Dict] = None
    ) -> None:
        """Set the default logging configuration for this factory.

        Args:
            dict_config: A logging dict config to utilize when adding new loggers. each logger
                added via add_logger method expects to find a handler called "default"
        """
        if dict_config is None:
            dict_config = default_config

        if not isinstance(dict_config, dict):
            raise ValueError(
                f"dict_config param must be a dict or None. Got {type(dict_config)}"
            )
        else:
            cls.dict_config = dict_config
            logging.config.dictConfig(cls.dict_config)

    @classmethod
    def attach_default_handler(
        cls: Type[JobmonLoggerConfig], logger_name: str, log_level: int = logging.INFO
    ) -> None:
        """A method to setup the default logging config for a specific logger_name.

        Args:
            logger_name: The logger to setup
            log_level: the log severity of the handler.

        Returns: None
        """
        logger_config = {
            "loggers": {
                logger_name: {
                    "handlers": ["default"],
                    "level": log_level,
                    "propagate": False,
                },
            }
        }
        cls.dict_config.update(logger_config)
        logging.config.dictConfig(cls.dict_config)



"""Tool represents a project or model that will be run many times over.

The Tool may evolve over time.
"""

from __future__ import annotations

import getpass
from http import HTTPStatus as StatusCodes
import logging
from typing import Any, Dict, List, Optional, Union

import yaml

from jobmon.client.task_template import TaskTemplate
from jobmon.client.tool_version import ToolVersion
from jobmon.client.workflow import Workflow
from jobmon.core.constants import MaxConcurrentlyRunning
from jobmon.core.exceptions import InvalidResponse
from jobmon.core.requester import Requester
from jobmon.core.serializers import SerializeClientTool

logger = logging.getLogger(__name__)


class InvalidToolError(Exception):
    """Exception for Tools that do not exist in the DB."""

    pass


class InvalidToolVersionError(Exception):
    """Exception for Tool version that is not valid."""

    pass


class Tool:
    """Tool represents a project or model that will be run many times over.

    The Tool may evolve over time.
    """

    def __init__(
        self,
        name: str = f"unknown-{getpass.getuser()}",
        active_tool_version_id: Union[str, int] = "latest",
        requester: Optional[Requester] = None,
    ) -> None:
        """A tool is an application which is expected to run many times on variable inputs.

         Which will serve a certain purpose over time even as the internal pipeline may change.
         Example tools are Dismod, Burdenator, Codem.

        Args:
            name: the name of the tool
            active_tool_version_id: which version of the tool to attach task templates and
                workflows to.
            requester: communicate with the flask services.
        """
        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

        # set tool defining attributes
        self.name = name
        self._bind()

        # import tool versions
        self.tool_versions = self._load_tool_versions()
        if not self.tool_versions:
            self.get_new_tool_version()
        else:
            self.set_active_tool_version_id(active_tool_version_id)

    def get_new_tool_version(self) -> int:
        """Create a new tool version for the current tool and activate it.

        Returns: the version id for the new tool
        """
        # call route to create tool version

        tool_version = ToolVersion.get_tool_version(tool=self)
        tool_version_id = tool_version.id
        self.tool_versions.append(tool_version)
        self.set_active_tool_version_id(tool_version_id)
        return tool_version_id

    @property
    def active_task_templates(self) -> Dict[str, TaskTemplate]:
        """Mapping of template_name to TaskTemplate for the active tool version."""
        return self.active_tool_version.task_templates

    @property
    def active_tool_version(self) -> ToolVersion:
        """Tool version id to use when spawning task templates."""
        return self._active_tool_version

    @property
    def default_compute_resources_set(self) -> Dict[str, Dict[str, Any]]:
        """Default compute resources associated with active tool version."""
        return self.active_tool_version.default_compute_resources_set

    @property
    def default_resource_scales_set(self) -> Dict[str, Dict[str, float]]:
        """Default resource scales associated with active tool version."""
        return self.active_tool_version.default_resource_scales_set

    @property
    def default_cluster_name(self) -> str:
        """Default cluster_name associated with active tool version."""
        return self.active_tool_version.default_cluster_name

    @property
    def default_max_attempts(self) -> Optional[int]:
        """Default max attempts of the active tool version."""
        return self.active_tool_version.default_max_attempt

    def set_active_tool_version_id(self, tool_version_id: Union[str, int]) -> None:
        """Tool version that is set as the active one (latest is default during instantiation).

        Args:
            tool_version_id: which tool version to set as active on this object.
        """
        version_index_lookup = {
            self.tool_versions[index].id: index
            for index in range(len(self.tool_versions))
        }

        # get the lookup value
        if tool_version_id == "latest":
            lookup_version: int = int(max(version_index_lookup.keys()))
        else:
            lookup_version = int(tool_version_id)

        # check that the version exists
        try:
            version_index = version_index_lookup[lookup_version]
        except KeyError:
            raise ValueError(
                f"{tool_version_id} is not a valid version for tool.name={self.name} Valid "
                f"versions={version_index_lookup.keys()}"
            )

        # set it as active and load task templates
        tool_version = self.tool_versions[version_index]
        tool_version.load_task_templates()
        self._active_tool_version: ToolVersion = tool_version

    def get_task_template(
        self,
        template_name: str,
        command_template: str,
        node_args: Optional[List[str]] = None,
        task_args: Optional[List[str]] = None,
        op_args: Optional[List[str]] = None,
        default_cluster_name: str = "",
        default_compute_resources: Optional[Dict[str, Any]] = None,
        default_resource_scales: Optional[Dict[str, float]] = None,
        yaml_file: Optional[str] = None,
        max_attempts: Optional[int] = None,
    ) -> TaskTemplate:
        """Create or get task a task template.

        Args:
            template_name: the name of this task template.
            command_template: an abstract command representing a task, where the arguments to
                the command have defined names but the values are not assigned. eg: '{python}
                {script} --data {data} --para {para} {verbose}'
            node_args: any named arguments in command_template that make the command unique
                within this template for a given workflow run. Generally these are arguments
                that can be parallelized over.
            task_args: any named arguments in command_template that make the command unique
                across workflows if the node args are the same as a previous workflow.
                Generally these are arguments about data moving though the task.
            op_args: any named arguments in command_template that can change without changing
                the identity of the task. Generally these are things like the task executable
                location or the verbosity of the script.
            default_cluster_name: the default cluster to run each task associated with this
                template on.
            default_compute_resources: dictionary of default compute resources to run tasks
                with. Can be overridden at task level. dict of {resource_name: resource_value}.
                Must specify default_cluster_name when this option is used.
            default_resource_scales: dictionary of default resource scales to adjust task
                resources with. Can be overridden at task level.
                dict of {resource_name: scale_factor}. Scale factor can be a numeric value, a
                Callable that will be applied to the existing resources, or an Iterator. Any
                Callable should take a single numeric value as its sole argument. Any
                Iterator should only yield numeric values. Any Iterable can be easily
                converted to an Iterator by using the iter() built-in (e.g. iter([80, 160,
                190])).
            yaml_file: path to YAML file that contains user-specified compute resources.
            max_attempts: max_attempts for the tt
        """
        if node_args is None:
            node_args = []
        if task_args is None:
            task_args = []
        if op_args is None:
            op_args = []

        if (
            default_compute_resources is not None or default_resource_scales is not None
        ) and not default_cluster_name:
            raise ValueError(
                "Must specify default_cluster_name when using "
                "default_compute_resources or default_resource_scales option"
            )

        tt = self.active_tool_version.get_task_template(template_name)

        # Read in compute resources and resources scales from YAML
        if yaml_file and (
            default_compute_resources is None or default_resource_scales is None
        ):
            with open(yaml_file, "r") as stream:
                try:
                    yaml_stream = yaml.safe_load(stream)
                except yaml.YAMLError as exc:
                    raise Exception(
                        f"Unable to read resources from {yaml_file}. "
                        f"Exception: {exc}"
                    )
            if default_compute_resources is None:
                default_compute_resources = yaml_stream["task_template_resources"][
                    tt.template_name
                ][default_cluster_name]
            if default_resource_scales is None:
                default_resource_scales = yaml_stream["task_template_scales"][
                    tt.template_name
                ][default_cluster_name]
        tt.get_task_template_version(
            command_template,
            node_args,
            task_args,
            op_args,
            default_max_attempts=max_attempts,
        )
        tt.default_cluster_name = default_cluster_name
        if default_compute_resources:
            tt.set_default_compute_resources_from_dict(
                default_cluster_name, default_compute_resources
            )
        if default_resource_scales:
            tt.set_default_resource_scales_from_dict(
                default_cluster_name, default_resource_scales
            )
        return tt

    def create_workflow(
        self,
        workflow_args: str = "",
        name: str = "",
        description: str = "",
        workflow_attributes: Optional[Union[List, dict]] = None,
        max_concurrently_running: int = MaxConcurrentlyRunning.MAXCONCURRENTLYRUNNING,
        chunk_size: int = 500,
        default_cluster_name: str = "",
        default_compute_resources_set: Optional[Dict] = None,
        default_resource_scales_set: Optional[Dict[str, float]] = None,
        default_max_attempts: Optional[int] = None,
    ) -> Workflow:
        """Create a workflow object associated with the active tool version.

        Args:
            workflow_args: Unique identifier of a workflow.
            name: Name of the workflow.
            description: Description of the workflow.
            workflow_attributes: Any key/value pair that the user wants to record for this
                workflow
            max_concurrently_running: How many running jobs to allow in parallel.
            chunk_size: how many tasks to bind in a single request
            default_cluster_name: name of cluster to run tasks on by default. Can be overridden
                at the task template or task level.
            default_compute_resources_set: dictionary of default compute resources to run tasks
                with. Can be overridden at task template or task level.
                dict of {cluster_name: {resource_name: resource_value}}
            default_resource_scales_set: dictionary of default resource_scales to adjust the
                resources with. Can be overridden at task template or task level.
                dict of {resource_name: scale_value}
            default_max_attempts: the default max_attempts value to use when create wf
        """
        wf = Workflow(
            self.active_tool_version,
            workflow_args,
            name,
            description,
            workflow_attributes,
            max_concurrently_running,
            requester=self.requester,
            chunk_size=chunk_size,
        )

        if default_max_attempts is None:
            default_max_attempts = self.default_max_attempts
        if default_max_attempts:
            wf.set_default_max_attempts(default_max_attempts)

        # set compute resource defaults
        if default_cluster_name:
            wf.default_cluster_name = default_cluster_name
        else:
            if self.default_cluster_name:
                wf.default_cluster_name = self.default_cluster_name
        if default_compute_resources_set:
            wf.default_compute_resources_set = default_compute_resources_set
        else:
            if self.active_tool_version.default_compute_resources_set:
                wf.default_compute_resources_set = self.default_compute_resources_set
        if default_resource_scales_set:
            wf.set_default_resource_scales_from_dict(
                cluster_name=default_cluster_name,
                dictionary=default_resource_scales_set,
            )
        else:
            if self.active_tool_version.default_resource_scales_set:
                wf.default_resource_scales_set = self.default_resource_scales_set

        return wf

    def update_default_compute_resources(
        self, cluster_name: str, **kwargs: Any
    ) -> None:
        """Update default compute resources in place only overridding specified keys.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to modify default values for.
            **kwargs: any key/value pair you want to update specified as an argument.
        """
        if not self.default_cluster_name:
            self.active_tool_version.default_cluster_name = cluster_name
        self.active_tool_version.update_default_compute_resources(
            cluster_name, **kwargs
        )

    def update_default_resource_scales(self, cluster_name: str, **kwargs: Any) -> None:
        """Update default resource scales in place only overridding specified keys.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to modify default values for.
            **kwargs: any key/value pair you want to update specified as an argument.
        """
        if not self.default_cluster_name:
            self.active_tool_version.default_cluster_name = cluster_name
        self.active_tool_version.update_default_resource_scales(cluster_name, **kwargs)

    def set_default_compute_resources_from_yaml(
        self,
        default_cluster_name: str,
        yaml_file: str,
        set_task_templates: bool = False,
        ignore_missing_keys: bool = False,
    ) -> None:
        """Set default compute resources from a user provided yaml file for tool level.

        Args:
            default_cluster_name: name of cluster to set default values for.
            yaml_file: the yaml file that is providing the default compute resource values.
            set_task_templates: whether or not the user wants to set the default compute
                resource values for all of the TaskTemplates associated with Tool.
            ignore_missing_keys: Whether or not to raise an error if a key is missing from the
                yaml file.
        """
        with open(yaml_file, "r") as stream:
            try:
                default_compute_resources = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                raise ValueError(
                    f"Unable to read default compute resources from " f"{yaml_file}."
                ) from exc

        # Set the Tool level compute resources
        try:
            compute_resources = default_compute_resources["tool_resources"][
                default_cluster_name
            ]
        except KeyError as exc:
            msg = f"No Tool resources matching cluster name in yaml file: {yaml_file}."
            if ignore_missing_keys:
                logger.info(msg)
            else:
                raise KeyError(msg) from exc

        self.active_tool_version.set_default_compute_resources_from_dict(
            cluster_name=default_cluster_name, compute_resources=compute_resources
        )
        self.active_tool_version.default_cluster_name = default_cluster_name

        if not set_task_templates:
            return

        if not self.active_task_templates:
            raise Exception(
                "No TaskTemplates associated with Tool, unable to set default "
                "compute resources for TaskTemplates."
            )

        if set_task_templates:
            # Set the compute resources for the TaskTemplates associated with the Tool
            for tt in self.active_task_templates.values():
                try:
                    tt.set_default_compute_resources_from_dict(
                        cluster_name=default_cluster_name,
                        compute_resources=(
                            default_compute_resources["task_template_resources"][
                                tt.template_name
                            ][default_cluster_name]
                        ),
                    )
                except KeyError as exc:
                    msg = (
                        f"No compute resources discovered in yaml file {yaml_file} for"
                        f"TaskTemplate {tt.template_name}"
                    )
                    if ignore_missing_keys:
                        logger.info(msg)
                    else:
                        raise KeyError(msg) from exc

    def set_default_resource_scales_from_yaml(
        self,
        default_cluster_name: str,
        yaml_file: str,
        set_task_templates: bool = False,
        ignore_missing_keys: bool = False,
    ) -> None:
        """Set default resource scales from a user provided yaml file for tool level.

        Args:
            default_cluster_name: name of cluster to set default values for.
            yaml_file: the yaml file that is providing the default compute resource values.
            set_task_templates: whether or not the user wants to set the default compute
                resource values for all of the TaskTemplates associated with Tool.
            ignore_missing_keys: Whether or not to raise an error if a key is missing from the
                yaml file.
        """
        with open(yaml_file, "r") as stream:
            try:
                default_resource_scales = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                raise ValueError(
                    f"Unable to read default resource scales from " f"{yaml_file}."
                ) from exc

        # Set the Tool level resource scales
        try:
            resource_scales = default_resource_scales["tool_scales"][
                default_cluster_name
            ]
        except KeyError as exc:
            msg = f"No Tool scales matching cluster name in yaml file: {yaml_file}."
            if ignore_missing_keys:
                logger.info(msg)
            else:
                raise KeyError(msg) from exc

        self.active_tool_version.set_default_resource_scales_from_dict(
            cluster_name=default_cluster_name, resource_scales=resource_scales
        )
        self.active_tool_version.default_cluster_name = default_cluster_name

        if not set_task_templates:
            return

        if not self.active_task_templates:
            raise Exception(
                "No TaskTemplates associated with Tool, unable to set default "
                "resource scales for TaskTemplates."
            )

        if set_task_templates:
            # Set the resource scales for the TaskTemplates associated with the Tool
            for tt in self.active_task_templates.values():
                try:
                    tt.set_default_resource_scales_from_dict(
                        cluster_name=default_cluster_name,
                        resource_scales=(
                            default_resource_scales["task_template_scales"][
                                tt.template_name
                            ][default_cluster_name]
                        ),
                    )
                except KeyError as exc:
                    msg = (
                        f"No resource scales discovered in yaml file {yaml_file} for"
                        f"TaskTemplate {tt.template_name}"
                    )
                    if ignore_missing_keys:
                        logger.info(msg)
                    else:
                        raise KeyError(msg) from exc

    def set_default_compute_resources_from_dict(
        self, cluster_name: str, compute_resources: Dict[str, Any]
    ) -> None:
        """Set default compute resources for a given cluster_name.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to set default values for.
            compute_resources: dictionary of default compute resources to run tasks
                with. Can be overridden at task template or task level.
                dict of {resource_name: resource_value}
        """
        if not self.default_cluster_name:
            self.active_tool_version.default_cluster_name = cluster_name
        self.active_tool_version.set_default_compute_resources_from_dict(
            cluster_name, compute_resources
        )

    def set_default_resource_scales_from_dict(
        self, cluster_name: str, resource_scales: Dict[str, float]
    ) -> None:
        """Set default compute resources for a given cluster_name.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to set default values for.
            resource_scales: dictionary of default resource scales to adjust task
                resources with. Can be overridden at task level.
                dict of {resource_name: scale_value}
        """
        if not self.default_cluster_name:
            self.active_tool_version.default_cluster_name = cluster_name
        self.active_tool_version.set_default_resource_scales_from_dict(
            cluster_name, resource_scales
        )

    def set_default_cluster_name(self, cluster_name: str) -> None:
        """Set default cluster.

        Args:
            cluster_name: name of cluster to set as default.
        """
        self.active_tool_version.default_cluster_name = cluster_name

    def set_default_max_attempts(self, value: int) -> None:
        """Set default max_attempts.

        Args:
            value: value of max_attempts.
        """
        self.active_tool_version.set_default_max_attempts(value)

    def set_default_clu(self, cluster_name: str) -> None:
        """Set default cluster.

        Args:
            cluster_name: name of cluster to set as default.
        """
        self.active_tool_version.default_cluster_name = cluster_name

    def _load_tool_versions(self) -> List[ToolVersion]:
        app_route = f"/tool/{self.id}/tool_versions"
        return_code, response = self.requester.send_request(
            app_route=app_route, message={}, request_type="get"
        )

        if return_code != StatusCodes.OK:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST request through route "
                f"{app_route}. Expected code 200. Response content: {response}"
            )

        tool_versions = [
            ToolVersion.from_wire(wire_tuple, self)
            for wire_tuple in response["tool_versions"]
        ]
        return tool_versions

    def _bind(self) -> None:
        """Call route to create tool."""
        app_route = "/tool"
        return_code, response = self.requester.send_request(
            app_route=app_route,
            message={"name": self.name},
            request_type="post",
        )

        if return_code != StatusCodes.OK:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST request through route "
                f"{app_route}. Expected code 200. Response content: {response}"
            )
        self.id = SerializeClientTool.kwargs_from_wire(response["tool"])["id"]

    def __repr__(self) -> str:
        """A representation string for a Tool instance."""
        return f"Tool(tool_id={self.id}, name={self.name}"
